from __future__ import annotations

from typing import TypeVar, Union

from sklearn.utils.validation import check_is_fitted

from ..._utils._sklearn_adapter import BaseEstimator, RegressorMixin
from ...misc.hat_matrix import HatMatrix, NadarayaWatsonHatMatrix
from ...misc.metrics import PairwiseMetric, l2_distance
from ...representation._functional_data import FData
from ...typing._metric import Metric
from ...typing._numpy import NDArrayFloat

Input = TypeVar("Input", bound=Union[NDArrayFloat, FData], contravariant=True)
Prediction = TypeVar("Prediction", bound=Union[NDArrayFloat, FData])


class KernelRegression(
    BaseEstimator,
    RegressorMixin[Input, Prediction],
):
    r"""Kernel regression with scalar response.

    Let :math:`fd_1 = (X_1, X_2, ..., X_n)` be the functional data set and
    :math:`y = (y_1, y_2, ..., y_n)` be the scalar response corresponding
    to each function in :math:`fd_1`. Then, the estimation for the
    functions in :math:`fd_2 = (X'_1, X'_2, ..., X'_n)` can be
    calculated as

    .. math::
        \hat{y} = \hat{H}y

    Where :math:`\hat{H}` is a matrix described in
    :class:`~skfda.misc.hat_matrix.HatMatrix`.

    Args:
        kernel_estimator: Method used to calculate the hat matrix
            (default =
            :class:`~skfda.misc.hat_matrix.NadarayaWatsonHatMatrix`).
        metric: Metric used to calculate the distances
            (default = :func:`L2 distance <skfda.misc.metrics.distance_l2>`).

    Examples:
        >>> import numpy as np
        >>> from skfda import FDataGrid
        >>> from skfda.misc.hat_matrix import NadarayaWatsonHatMatrix
        >>> from skfda.misc.hat_matrix import KNeighborsHatMatrix

        >>> grid_points = np.linspace(0, 1, num=11)
        >>> data1 = np.array([i + grid_points for i in range(1, 9, 2)])
        >>> data2 = np.array([i + grid_points for i in range(2, 7, 2)])

        >>> fd_1 = FDataGrid(grid_points=grid_points, data_matrix=data1)
        >>> y = np.array([1, 3, 5, 7])
        >>> fd_2 = FDataGrid(grid_points=grid_points, data_matrix=data2)

        >>> kernel_estimator = NadarayaWatsonHatMatrix(bandwidth=1)
        >>> estimator = KernelRegression(kernel_estimator=kernel_estimator)
        >>> _ = estimator.fit(fd_1, y)
        >>> estimator.predict(fd_2)
        array([ 2.02723928,  4.        ,  5.97276072])

        >>> kernel_estimator = KNeighborsHatMatrix(n_neighbors=2)
        >>> estimator = KernelRegression(kernel_estimator=kernel_estimator)
        >>> _ = estimator.fit(fd_1, y)
        >>> estimator.predict(fd_2)
        array([ 2.,  4.,  6.])

    """

    def __init__(
        self,
        *,
        kernel_estimator: HatMatrix | None = None,
        metric: Metric[Input] = l2_distance,
    ):

        self.kernel_estimator = kernel_estimator
        self.metric = metric

    def fit(  # noqa: D102
        self,
        X: Input,
        y: Prediction,
        weight: NDArrayFloat | None = None,
    ) -> KernelRegression[Input, Prediction]:

        self.X_train_ = X
        self.y_train_ = y
        self.weights_ = weight

        self._kernel_estimator: HatMatrix
        if self.kernel_estimator is None:
            self._kernel_estimator = NadarayaWatsonHatMatrix()
        else:
            self._kernel_estimator = self.kernel_estimator

        return self

    def predict(  # noqa: D102
        self,
        X: Input,
    ) -> Prediction:

        check_is_fitted(self)
        delta_x = PairwiseMetric(self.metric)(X, self.X_train_)

        return self._kernel_estimator(
            delta_x=delta_x,
            X_train=self.X_train_,
            y_train=self.y_train_,
            X=X,
            weights=self.weights_,
        )



"""Neighbor models for regression."""

from __future__ import annotations

from typing import Tuple, TypeVar, Union, overload

from sklearn.neighbors import (
    KNeighborsRegressor as _KNeighborsRegressor,
    RadiusNeighborsRegressor as _RadiusNeighborsRegressor,
)
from typing_extensions import Literal

from ..._utils._neighbors_base import (
    AlgorithmType,
    KNeighborsMixin,
    NeighborsRegressorMixin,
    RadiusNeighborsMixin,
    WeightsType,
)
from ...misc.metrics import l2_distance
from ...representation import FData
from ...typing._metric import Metric
from ...typing._numpy import NDArrayFloat, NDArrayInt

InputBound = Union[NDArrayFloat, FData]
Input = TypeVar("Input", contravariant=True, bound=InputBound)
TargetBound = Union[NDArrayFloat, FData]
Target = TypeVar("Target", bound=TargetBound)


class KNeighborsRegressor(
    NeighborsRegressorMixin[Input, Target],
    KNeighborsMixin[Input, Target],
):
    """
    Regression based on k-nearest neighbors.

    Regression with scalar, multivariate or functional response.

    The target is predicted by local interpolation of the targets associated of
    the nearest neighbors in the training set.

    Parameters:
        n_neighbors: Number of neighbors to use by default for
            :meth:`kneighbors` queries.
        weights: Weight function used in prediction.  Possible values:

            - 'uniform' : uniform weights.  All points in each neighborhood
              are weighted equally.
            - 'distance' : weight points by the inverse of their distance.
              in this case, closer neighbors of a query point will have a
              greater influence than neighbors which are further away.
            - [callable] : a user-defined function which accepts an
              array of distances, and returns an array of the same shape
              containing the weights.

        algorithm: Algorithm used to compute the nearest neighbors:

            - 'ball_tree' will use :class:`sklearn.neighbors.BallTree`.
            - 'brute' will use a brute-force search.
            - 'auto' will attempt to decide the most appropriate algorithm
              based on the values passed to :meth:`fit` method.

        leaf_size: Leaf size passed to BallTree or KDTree. This can affect the
            speed of the construction and query, as well as the memory
            required to store the tree. The optimal value depends on the
            nature of the problem.
        metric: The distance metric to use for the tree.  The default metric is
            the L2 distance. See the documentation of the metrics module
            for a list of available metrics.
        n_jobs: The number of parallel jobs to run for neighbors search.
            ``None`` means 1 unless in a :obj:`joblib.parallel_backend`
            context.
            ``-1`` means using all processors.
            Doesn't affect :meth:`fit` method.

    Examples:
        Firstly, we will create a toy dataset with gaussian-like samples
        shifted.

        >>> from skfda.ml.regression import KNeighborsRegressor
        >>> from skfda.datasets import make_multimodal_samples
        >>> from skfda.datasets import make_multimodal_landmarks
        >>> y = make_multimodal_landmarks(
        ...     n_samples=30,
        ...     std=0.5,
        ...     random_state=0,
        ... )
        >>> y_train = y.flatten()
        >>> X_train = make_multimodal_samples(
        ...     n_samples=30,
        ...     std=0.5,
        ...     random_state=0,
        ... )
        >>> X_test = make_multimodal_samples(
        ...     n_samples=5,
        ...     std=0.05,
        ...     random_state=0,
        ... )

        We will fit a K-Nearest Neighbors regressor to regress a scalar
        response.

        >>> neigh = KNeighborsRegressor()
        >>> neigh.fit(X_train, y_train)
        KNeighborsRegressor(...)

        We can predict the modes of new samples

        >>> neigh.predict(X_test).round(2) # Predict test data
        array([ 0.38, 0.14, 0.27, 0.52, 0.38])


        Now we will create a functional response to train the model

        >>> y_train = 5 * X_train + 1
        >>> y_train
        FDataGrid(...)

        We train the estimator with the functional response

        >>> neigh.fit(X_train, y_train)
        KNeighborsRegressor(...)

        And predict the responses as in the first case.

        >>> neigh.predict(X_test)
        FDataGrid(...)

    See also:
        :class:`~skfda.ml.classification.KNeighborsClassifier`
        :class:`~skfda.ml.classification.RadiusNeighborsClassifier`
        :class:`~skfda.ml.classification.NearestCentroids`
        :class:`~skfda.ml.regression.RadiusNeighborsRegressor`
        :class:`~skfda.ml.clustering.NearestNeighbors`


    Notes:
        See Nearest Neighbors in the sklearn online documentation for a
        discussion of the choice of ``algorithm`` and ``leaf_size``.

        This class wraps the sklearn regressor
        `sklearn.neighbors.KNeighborsRegressor`.

        .. warning::
           Regarding the Nearest Neighbors algorithms, if it is found that two
           neighbors, neighbor `k+1` and `k`, have identical distances
           but different labels, the results will depend on the ordering of the
           training data.

        https://en.wikipedia.org/wiki/K-nearest_neighbor_algorithm

    """

    @overload
    def __init__(
        self: KNeighborsRegressor[NDArrayFloat, Target],
        *,
        n_neighbors: int = 5,
        weights: WeightsType = 'uniform',
        algorithm: AlgorithmType = 'auto',
        leaf_size: int = 30,
        metric: Literal["precomputed"],
        n_jobs: int | None = None,
    ) -> None:
        pass

    @overload
    def __init__(
        self: KNeighborsRegressor[InputBound, Target],
        *,
        n_neighbors: int = 5,
        weights: WeightsType = 'uniform',
        algorithm: AlgorithmType = 'auto',
        leaf_size: int = 30,
        n_jobs: int | None = None,
    ) -> None:
        pass

    @overload
    def __init__(
        self,
        *,
        n_neighbors: int = 5,
        weights: WeightsType = 'uniform',
        algorithm: AlgorithmType = 'auto',
        leaf_size: int = 30,
        metric: Metric[Input] = l2_distance,
        n_jobs: int | None = None,
    ) -> None:
        pass

    # Not useless: it restrict the inputs.
    def __init__(  # noqa: WPS612
        self,
        *,
        n_neighbors: int = 5,
        weights: WeightsType = 'uniform',
        algorithm: AlgorithmType = 'auto',
        leaf_size: int = 30,
        metric: Literal["precomputed"] | Metric[Input] = l2_distance,
        n_jobs: int | None = None,
    ) -> None:
        """Initialize the regressor."""
        super().__init__(
            n_neighbors=n_neighbors,
            weights=weights,
            algorithm=algorithm,
            leaf_size=leaf_size,
            metric=metric,
            n_jobs=n_jobs,
        )

    def _init_estimator(self) -> _KNeighborsRegressor:
        return _KNeighborsRegressor(
            n_neighbors=self.n_neighbors,
            weights=self.weights,
            algorithm=self.algorithm,
            leaf_size=self.leaf_size,
            metric="precomputed",
            n_jobs=self.n_jobs,
        )

    def _query(
        self,
        X: Input,
    ) -> Tuple[NDArrayFloat, NDArrayInt]:
        """Return distances and neighbors of given sample."""
        return self.kneighbors(X)


class RadiusNeighborsRegressor(
    NeighborsRegressorMixin[Input, Target],
    RadiusNeighborsMixin[Input, Target],
):
    """
    Regression based on neighbors within a fixed radius.

    Regression with scalar, multivariate or functional response.

    The target is predicted by local interpolation of the targets associated of
    the nearest neighbors in the training set.

    Parameters:
        radius: Range of parameter space to use by default for
            :meth:`radius_neighbors` queries.
        weights: Weight function used in prediction.  Possible values:

            - 'uniform' : uniform weights.  All points in each neighborhood
              are weighted equally.
            - 'distance' : weight points by the inverse of their distance.
              in this case, closer neighbors of a query point will have a
              greater influence than neighbors which are further away.
            - [callable] : a user-defined function which accepts an
              array of distances, and returns an array of the same shape
              containing the weights.

            Uniform weights are used by default.
        algorithm: Algorithm used to compute the nearest neighbors:

            - 'ball_tree' will use :class:`sklearn.neighbors.BallTree`.
            - 'brute' will use a brute-force search.
            - 'auto' will attempt to decide the most appropriate algorithm
              based on the values passed to :meth:`fit` method.

        leaf_size: Leaf size passed to BallTree. This can affect the
            speed of the construction and query, as well as the memory
            required to store the tree. The optimal value depends on the
            nature of the problem.
        metric: The distance metric to use for the tree.  The default metric is
            the L2 distance. See the documentation of the metrics module
            for a list of available metrics.
        n_jobs: The number of parallel jobs to run for neighbors search.
            ``None`` means 1 unless in a :obj:`joblib.parallel_backend`
            context.
            ``-1`` means using all processors.

    Examples:
        Firstly, we will create a toy dataset with gaussian-like samples
        shifted.

        >>> from skfda.ml.regression import RadiusNeighborsRegressor
        >>> from skfda.datasets import make_multimodal_samples
        >>> from skfda.datasets import make_multimodal_landmarks
        >>> y = make_multimodal_landmarks(
        ...     n_samples=30,
        ...     std=0.5,
        ...     random_state=0,
        ... )
        >>> y_train = y.flatten()
        >>> X_train = make_multimodal_samples(
        ...     n_samples=30,
        ...     std=0.5,
        ...     random_state=0,
        ... )
        >>> X_test = make_multimodal_samples(
        ...     n_samples=5,
        ...     std=0.05,
        ...     random_state=0,
        ... )

        We will fit a Radius-Nearest Neighbors regressor to regress a scalar
        response.

        >>> neigh = RadiusNeighborsRegressor(radius=0.2)
        >>> neigh.fit(X_train, y_train)
        RadiusNeighborsRegressor(...radius=0.2...)

        We can predict the modes of new samples

        >>> neigh.predict(X_test).round(2) # Predict test data
        array([ 0.39, 0.07, 0.26, 0.5 , 0.46])


        Now we will create a functional response to train the model

        >>> y_train = 5 * X_train + 1
        >>> y_train
        FDataGrid(...)

        We train the estimator with the functional response

        >>> neigh.fit(X_train, y_train)
        RadiusNeighborsRegressor(...radius=0.2...)

        And predict the responses as in the first case.

        >>> neigh.predict(X_test)
        FDataGrid(...)

    See also:
        :class:`~skfda.ml.classification.KNeighborsClassifier`
        :class:`~skfda.ml.classification.RadiusNeighborsClassifier`
        :class:`~skfda.ml.classification.NearestCentroids`
        :class:`~skfda.ml.regression.KNeighborsRegressor`
        :class:`~skfda.ml.clustering.NearestNeighbors`


    Notes:
        See Nearest Neighbors in the sklearn online documentation for a
        discussion of the choice of ``algorithm`` and ``leaf_size``.

        This class wraps the sklearn classifier
        `sklearn.neighbors.RadiusNeighborsClassifier`.

        https://en.wikipedia.org/wiki/K-nearest_neighbor_algorithm

    """

    @overload
    def __init__(
        self: RadiusNeighborsRegressor[NDArrayFloat, Target],
        *,
        radius: float = 1.0,
        weights: WeightsType = 'uniform',
        algorithm: AlgorithmType = 'auto',
        leaf_size: int = 30,
        metric: Literal["precomputed"],
        n_jobs: int | None = None,
    ) -> None:
        pass

    @overload
    def __init__(
        self: RadiusNeighborsRegressor[InputBound, Target],
        *,
        radius: float = 1.0,
        weights: WeightsType = 'uniform',
        algorithm: AlgorithmType = 'auto',
        leaf_size: int = 30,
        n_jobs: int | None = None,
    ) -> None:
        pass

    @overload
    def __init__(
        self,
        *,
        radius: float = 1.0,
        weights: WeightsType = 'uniform',
        algorithm: AlgorithmType = 'auto',
        leaf_size: int = 30,
        metric: Metric[Input] = l2_distance,
        n_jobs: int | None = None,
    ) -> None:
        pass

    # Parameters are important
    def __init__(  # noqa: WPS612
        self,
        *,
        radius: float = 1.0,
        weights: WeightsType = 'uniform',
        algorithm: AlgorithmType = 'auto',
        leaf_size: int = 30,
        metric: Literal["precomputed"] | Metric[Input] = l2_distance,
        n_jobs: int | None = None,
    ) -> None:
        """Initialize the classifier."""
        super().__init__(
            radius=radius,
            weights=weights,
            algorithm=algorithm,
            leaf_size=leaf_size,
            metric=metric,
            n_jobs=n_jobs,
        )

    def _init_estimator(self) -> _RadiusNeighborsRegressor:
        return _RadiusNeighborsRegressor(
            radius=self.radius,
            weights=self.weights,
            algorithm=self.algorithm,
            leaf_size=self.leaf_size,
            metric="precomputed",
            n_jobs=self.n_jobs,
        )

    def _query(
        self,
        X: Input,
    ) -> Tuple[NDArrayFloat, NDArrayInt]:
        return self.radius_neighbors(X)



"""Test to check the per functional transformers."""

import unittest

import numpy as np

import skfda.representation.basis as basis
from skfda.datasets import fetch_growth
from skfda.preprocessing.feature_construction import (
    unconditional_expected_value,
)


class TestUnconditionalExpectedValues(unittest.TestCase):
    """Tests for unconditional expected values method."""

    def test_transform(self) -> None:
        """Compare results with grid and basis representations."""
        X = fetch_growth(return_X_y=True)[0]

        data_grid = unconditional_expected_value(
            X[:5],
            np.log,
        )
        data_basis = unconditional_expected_value(
            X[:5].to_basis(basis.BSplineBasis(n_basis=7)),
            np.log,
        )
        np.testing.assert_allclose(data_basis, data_grid, rtol=1e-3)


if __name__ == '__main__':
    unittest.main()



"""Variable selection methods for functional data."""
from typing import TYPE_CHECKING

import lazy_loader as lazy

__getattr__, __dir__, __all__ = lazy.attach(
    __name__,
    submodules=[
        "maxima_hunting",
        "mrmr",
        "recursive_maxima_hunting",
    ],
    submod_attrs={
        "_rkvs": ["RKHSVariableSelection"],
        "maxima_hunting": ["MaximaHunting"],
        "mrmr": ["MinimumRedundancyMaximumRelevance"],
        "recursive_maxima_hunting": ["RecursiveMaximaHunting"],
    },
)

if TYPE_CHECKING:
    from ._rkvs import RKHSVariableSelection as RKHSVariableSelection
    from .maxima_hunting import MaximaHunting as MaximaHunting
    from .mrmr import (
        MinimumRedundancyMaximumRelevance as MinimumRedundancyMaximumRelevance,
    )
    from .recursive_maxima_hunting import (
        RecursiveMaximaHunting as RecursiveMaximaHunting,
    )



"""Test kernel regression method."""
import unittest
from typing import Callable, Optional, Tuple

import numpy as np
import sklearn.model_selection

from skfda import FData
from skfda.datasets import fetch_tecator
from skfda.misc.hat_matrix import (
    KNeighborsHatMatrix,
    LocalLinearRegressionHatMatrix,
    NadarayaWatsonHatMatrix,
)
from skfda.misc.kernels import normal, uniform
from skfda.misc.metrics import l2_distance
from skfda.ml.regression import KernelRegression
from skfda.representation.basis import FDataBasis, FourierBasis, MonomialBasis
from skfda.representation.grid import FDataGrid

FloatArray = np.typing.NDArray[np.float_]


def _nw_alt(
    fd_train: FData,
    fd_test: FData,
    y_train: FloatArray,
    *,
    bandwidth: float,
    kernel: Optional[Callable[[FloatArray], FloatArray]] = None,
) -> FloatArray:
    if kernel is None:
        kernel = normal

    y = np.zeros(fd_test.n_samples)
    for i in range(fd_test.n_samples):
        w = kernel(l2_distance(fd_train, fd_test[i]) / bandwidth)
        y[i] = (w @ y_train) / sum(w)

    return y


def _knn_alt(
    fd_train: FData,
    fd_test: FData,
    y_train: FloatArray,
    *,
    bandwidth: int,
    kernel: Optional[Callable[[FloatArray], FloatArray]] = None,
) -> FloatArray:
    if kernel is None:
        kernel = uniform

    y = np.zeros(fd_test.n_samples)

    for i in range(fd_test.n_samples):
        d = l2_distance(fd_train, fd_test[i])
        tol = np.finfo(np.float64).eps
        h = sorted(d)[bandwidth - 1] + tol
        w = kernel(d / h)
        y[i] = (w @ y_train) / sum(w)

    return y


def _llr_alt(
    fd_train: FDataBasis,
    fd_test: FDataBasis,
    y_train: FloatArray,
    *,
    bandwidth: float,
    kernel: Optional[Callable[[FloatArray], FloatArray]] = None,
) -> FloatArray:
    if kernel is None:
        kernel = normal

    y = np.zeros(fd_test.n_samples)

    for i in range(fd_test.n_samples):
        d = l2_distance(fd_train, fd_test[i])
        W = np.diag(kernel(d / bandwidth))

        C = np.concatenate(
            (
                np.ones(fd_train.n_samples)[:, np.newaxis],
                (fd_train - fd_test[i]).coefficients,
            ),
            axis=1,
        )

        M = np.linalg.inv(np.linalg.multi_dot([C.T, W, C]))
        y[i] = np.linalg.multi_dot([M, C.T, W, y_train])[0]

    return y


def _create_data_basis(
) -> Tuple[FDataBasis, FDataBasis, FloatArray]:
    X, y = fetch_tecator(return_X_y=True, as_frame=True)
    fd = X.iloc[:, 0].values
    fat = y['fat'].values

    basis = FourierBasis(
        n_basis=10,
        domain_range=fd.domain_range,
    )

    fd_basis = fd.to_basis(basis=basis)

    fd_train, fd_test, y_train, _ = sklearn.model_selection.train_test_split(
        fd_basis,
        fat,
        test_size=0.2,
        random_state=10,
    )
    return fd_train, fd_test, y_train


def _create_data_grid(
) -> Tuple[FDataGrid, FDataGrid, FloatArray]:
    X, y = fetch_tecator(return_X_y=True, as_frame=True)
    fd = X.iloc[:, 0].values
    fat = y['fat'].values

    fd_train, fd_test, y_train, _ = sklearn.model_selection.train_test_split(
        fd,
        fat,
        test_size=0.2,
        random_state=10,
    )

    return fd_train, fd_test, y_train


def _create_data_r(
) -> Tuple[FDataGrid, FDataGrid, FloatArray]:
    X, y = fetch_tecator(return_X_y=True, as_frame=True)
    fd = X.iloc[:, 0].values
    fat = y['fat'].values

    return fd[:100], fd[100:110], fat[:100]


class TestKernelRegression(unittest.TestCase):
    """Test Nadaraya-Watson, KNNeighbours and LocalLinearRegression methods."""

    def test_nadaraya_watson(self) -> None:
        """Test Nadaraya-Watson method."""
        # Creating data
        fd_train_basis, fd_test_basis, y_train_basis = _create_data_basis()
        fd_train_grid, fd_test_grid, y_train_grid = _create_data_grid()

        # Test NW method with basis representation and bandwidth=1
        nw_basis = KernelRegression[FDataBasis, np.typing.NDArray[np.float_]](
            kernel_estimator=NadarayaWatsonHatMatrix(bandwidth=1),
        )
        nw_basis.fit(fd_train_basis, y_train_basis)
        y_basis = nw_basis.predict(fd_test_basis)

        np.testing.assert_allclose(
            _nw_alt(
                fd_train_basis,
                fd_test_basis,
                y_train_basis,
                bandwidth=1,
            ),
            y_basis,
        )

        # Test NW method with grid representation and bandwidth=1
        nw_grid = KernelRegression[FDataGrid, np.typing.NDArray[np.float_]](
            kernel_estimator=NadarayaWatsonHatMatrix(bandwidth=1),
        )
        nw_grid.fit(fd_train_grid, y_train_grid)
        y_grid = nw_grid.predict(fd_test_grid)

        np.testing.assert_allclose(
            _nw_alt(
                fd_train_grid,
                fd_test_grid,
                y_train_grid,
                bandwidth=1,
            ),
            y_grid,
        )

    def test_knn(self) -> None:
        """Test K-Nearest Neighbours method."""
        # Creating data
        fd_train_basis, fd_test_basis, y_train_basis = _create_data_basis()
        fd_train_grid, fd_test_grid, y_train_grid = _create_data_grid()

        # Test KNN method with basis representation, n_neighbours=3 and
        # uniform kernel
        knn_basis = KernelRegression[FDataBasis, np.typing.NDArray[np.float_]](
            kernel_estimator=KNeighborsHatMatrix(n_neighbors=3),
        )
        knn_basis.fit(fd_train_basis, y_train_basis)
        y_basis = knn_basis.predict(fd_test_basis)

        np.testing.assert_allclose(
            _knn_alt(
                fd_train_basis,
                fd_test_basis,
                y_train_basis,
                bandwidth=3,
            ),
            y_basis,
        )

        # Test KNN method with grid representation, n_neighbours=3 and
        # uniform kernel
        knn_grid = KernelRegression[FDataGrid, np.typing.NDArray[np.float_]](
            kernel_estimator=KNeighborsHatMatrix(n_neighbors=3),
        )
        knn_grid.fit(fd_train_grid, y_train_grid)
        y_grid = knn_grid.predict(fd_test_grid)

        np.testing.assert_allclose(
            _knn_alt(
                fd_train_grid,
                fd_test_grid,
                y_train_grid,
                bandwidth=3,
            ),
            y_grid,
        )

        # Test KNN method with basis representation, n_neighbours=10 and
        # normal kernel
        knn_basis = KernelRegression(
            kernel_estimator=KNeighborsHatMatrix(
                n_neighbors=10,
                kernel=normal,
            ),
        )
        knn_basis.fit(fd_train_basis, y_train_basis)
        y_basis = knn_basis.predict(fd_test_basis)

        np.testing.assert_allclose(
            _knn_alt(
                fd_train_basis,
                fd_test_basis,
                y_train_basis,
                bandwidth=10,
                kernel=normal,
            ),
            y_basis,
        )

    def test_llr(self) -> None:
        """Test Local Linear Regression method."""
        # Creating data
        fd_train_basis, fd_test_basis, y_train_basis = _create_data_basis()

        llr_basis = KernelRegression[FDataBasis, np.typing.NDArray[np.float_]](
            kernel_estimator=LocalLinearRegressionHatMatrix(bandwidth=1),
        )
        llr_basis.fit(fd_train_basis, y_train_basis)
        y_basis = llr_basis.predict(fd_test_basis)

        np.testing.assert_allclose(
            _llr_alt(
                fd_train_basis,
                fd_test_basis,
                y_train_basis,
                bandwidth=1,
            ),
            y_basis,
        )

    def test_nw_r(self) -> None:
        """Comparison of NW's results with results from fda.usc."""
        X_train, X_test, y_train = _create_data_r()

        nw = KernelRegression[FDataGrid, np.typing.NDArray[np.float_]](
            kernel_estimator=NadarayaWatsonHatMatrix(bandwidth=1),
        )
        nw.fit(X_train, y_train)

        y = nw.predict(X_test)
        result_R = [
            18.245093,
            22.976695,
            9.429236,
            16.852003,
            16.568529,
            8.520466,
            14.943808,
            15.344949,
            8.646862,
            16.576900,
        ]

        np.testing.assert_almost_equal(y, result_R, decimal=3)

    def test_knn_r(self) -> None:
        """Comparison of NW's results with results from fda.usc."""
        X_train, X_test, y_train = _create_data_r()

        knn = KernelRegression[FDataGrid, np.typing.NDArray[np.float_]](
            kernel_estimator=KNeighborsHatMatrix(n_neighbors=3),
        )
        knn.fit(X_train, y_train)

        y = knn.predict(X_test)
        result_R = [
            20.400000,
            24.166667,
            10.900000,
            20.466667,
            16.900000,
            5.433333,
            14.400000,
            11.966667,
            9.033333,
            19.633333,
        ]

        np.testing.assert_almost_equal(y, result_R, decimal=6)


class TestNonOthonormalBasisLLR(unittest.TestCase):
    """Test LocalLinearRegression method with non orthonormal basis."""

    def test_llr_non_orthonormal(self) -> None:
        """Test LocalLinearRegression with monomial basis."""
        coef1 = [[1, 5, 8], [4, 6, 6], [9, 4, 1]]
        coef2 = [[6, 3, 5]]
        basis = MonomialBasis(n_basis=3, domain_range=(0, 3))

        X_train = FDataBasis(coefficients=coef1, basis=basis)
        X = FDataBasis(coefficients=coef2, basis=basis)
        y_train = np.array([8, 6, 1])

        llr = LocalLinearRegressionHatMatrix(
            bandwidth=100,
            kernel=uniform,
        )
        kr = KernelRegression[FDataBasis, np.typing.NDArray[np.float_]](
            kernel_estimator=llr,
        )
        kr.fit(X_train, y_train)
        np.testing.assert_almost_equal(kr.predict(X), 4.35735166)



from __future__ import annotations

from typing import Any, Literal, TypeVar, Union, overload

from scipy.sparse import csr_matrix
from sklearn.neighbors import KNeighborsTransformer as _KNeighborsTransformer

from skfda._utils._sklearn_adapter import InductiveTransformerMixin

from ..._utils._neighbors_base import AlgorithmType, KNeighborsMixin
from ...misc.metrics import l2_distance
from ...representation import FData
from ...typing._metric import Metric
from ...typing._numpy import NDArrayFloat

InputBound = Union[NDArrayFloat, FData]
Input = TypeVar("Input", contravariant=True, bound=InputBound)
Target = TypeVar("Target")
SelfType = TypeVar("SelfType", bound="KNeighborsTransformer[Any, Any]")


class KNeighborsTransformer(
    KNeighborsMixin[Input, Any],
    InductiveTransformerMixin[Input, NDArrayFloat, Any],
):

    n_neighbors: int

    @overload
    def __init__(
        self: KNeighborsTransformer[NDArrayFloat],
        *,
        mode: Literal["connectivity", "distance"] = "distance",
        n_neighbors: int = 5,
        algorithm: AlgorithmType = 'auto',
        leaf_size: int = 30,
        metric: Literal["precomputed"],
        n_jobs: int | None = None,
    ) -> None:
        pass

    @overload
    def __init__(
        self: KNeighborsTransformer[InputBound],
        *,
        mode: Literal["connectivity", "distance"] = "distance",
        n_neighbors: int = 5,
        algorithm: AlgorithmType = 'auto',
        leaf_size: int = 30,
        n_jobs: int | None = None,
    ) -> None:
        pass

    @overload
    def __init__(
        self,
        *,
        mode: Literal["connectivity", "distance"] = "distance",
        n_neighbors: int = 5,
        algorithm: AlgorithmType = 'auto',
        leaf_size: int = 30,
        metric: Metric[Input] = l2_distance,
        n_jobs: int | None = None,
    ) -> None:
        pass

    # Not useless, it restricts parameters
    def __init__(  # noqa: WPS612
        self,
        *,
        mode: Literal["connectivity", "distance"] = "distance",
        n_neighbors: int = 5,
        algorithm: AlgorithmType = 'auto',
        leaf_size: int = 30,
        metric: Literal["precomputed"] | Metric[Input] = l2_distance,
        n_jobs: int | None = None,
    ) -> None:
        self.mode = mode
        super().__init__(
            n_neighbors=n_neighbors,
            algorithm=algorithm,
            leaf_size=leaf_size,
            metric=metric,
            n_jobs=n_jobs,
        )

    def _init_estimator(self) -> _KNeighborsTransformer:

        return _KNeighborsTransformer(
            mode=self.mode,
            n_neighbors=self.n_neighbors,
            algorithm=self.algorithm,
            leaf_size=self.leaf_size,
            metric="precomputed",
            n_jobs=self.n_jobs,
        )

    def _fit(
        self: SelfType,
        X: Input,
        y: Target,
        fit_with_zeros: bool = True,
    ) -> SelfType:
        ret = super()._fit(X, y)

        self.n_features_in_ = 1 if isinstance(X, FData) else X.shape[1]

        return ret

    def transform(
        self,
        X: Input,
    ) -> csr_matrix:
        self._check_is_fitted()
        add_one = self.mode == "distance"
        return self.kneighbors_graph(
            X,
            mode=self.mode,
            n_neighbors=self.n_neighbors + add_one,
        )



# -*- coding: utf-8 -*-

"""Kernel Smoother.

This module contains the class for kernel smoothing.

"""
from typing import Optional

import numpy as np

from ..._utils._utils import _cartesian_product, _to_grid_points
from ...misc.hat_matrix import HatMatrix, NadarayaWatsonHatMatrix
from ...misc.metrics import PairwiseMetric, l2_distance
from ...typing._base import GridPointsLike
from ...typing._metric import Metric
from ...typing._numpy import NDArrayFloat
from ._linear import _LinearSmoother


class KernelSmoother(_LinearSmoother):
    r"""Kernel smoothing method.

    This module allows to perform functional data smoothing.

    Let :math:`t = (t_1, t_2, ..., t_n)` be the
    points of discretisation and :math:`X` the vector of observations at that
    points. Then, the smoothed values, :math:`\hat{X}`, at the points
    :math:`t' = (t_1', t_2', ..., t_m')` are obtained as

    .. math::
        \hat{X} = \hat{H} X

    where :math:`\hat{H}` is a matrix described in
    :class:`~skfda.misc.hat_matrix.HatMatrix`.

    Examples:
        >>> from skfda import FDataGrid
        >>> from skfda.misc.hat_matrix import NadarayaWatsonHatMatrix
        >>> fd = FDataGrid(
        ...     grid_points=[1, 2, 4, 5, 7],
        ...     data_matrix=[[1, 2, 3, 4, 5]],
        ... )
        >>> kernel_estimator = NadarayaWatsonHatMatrix(bandwidth=3.5)
        >>> smoother = KernelSmoother(kernel_estimator=kernel_estimator)
        >>> fd_smoothed = smoother.fit_transform(fd)
        >>> fd_smoothed.data_matrix.round(2)
        array([[[ 2.42],
                [ 2.61],
                [ 3.03],
                [ 3.24],
                [ 3.65]]])
        >>> smoother.hat_matrix().round(3)
        array([[ 0.294, 0.282, 0.204, 0.153, 0.068],
               [ 0.249, 0.259, 0.22 , 0.179, 0.093],
               [ 0.165, 0.202, 0.238, 0.229, 0.165],
               [ 0.129, 0.172, 0.239, 0.249, 0.211],
               [ 0.073, 0.115, 0.221, 0.271, 0.319]])
        >>> kernel_estimator = NadarayaWatsonHatMatrix(bandwidth=2)
        >>> smoother = KernelSmoother(kernel_estimator=kernel_estimator)
        >>> fd_smoothed = smoother.fit_transform(fd)
        >>> fd_smoothed.data_matrix.round(2)
        array([[[ 1.84],
                [ 2.18],
                [ 3.09],
                [ 3.55],
                [ 4.28]]])
        >>> smoother.hat_matrix().round(3)
        array([[ 0.425, 0.375, 0.138, 0.058, 0.005],
               [ 0.309, 0.35 , 0.212, 0.114, 0.015],
               [ 0.103, 0.193, 0.319, 0.281, 0.103],
               [ 0.046, 0.11 , 0.299, 0.339, 0.206],
               [ 0.006, 0.022, 0.163, 0.305, 0.503]])

        The output points can be changed:

        >>> kernel_estimator = NadarayaWatsonHatMatrix(bandwidth=2)
        >>> smoother = KernelSmoother(
        ...     kernel_estimator=kernel_estimator,
        ...     output_points=[1, 2, 3, 4, 5, 6, 7],
        ... )
        >>> fd_smoothed = smoother.fit_transform(fd)
        >>> fd_smoothed.data_matrix.round(2)
        array([[[ 1.84],
                [ 2.18],
                [ 2.61],
                [ 3.09],
                [ 3.55],
                [ 3.95],
                [ 4.28]]])
        >>> smoother.hat_matrix().round(3)
        array([[ 0.425,  0.375,  0.138,  0.058,  0.005],
               [ 0.309,  0.35 ,  0.212,  0.114,  0.015],
               [ 0.195,  0.283,  0.283,  0.195,  0.043],
               [ 0.103,  0.193,  0.319,  0.281,  0.103],
               [ 0.046,  0.11 ,  0.299,  0.339,  0.206],
               [ 0.017,  0.053,  0.238,  0.346,  0.346],
               [ 0.006,  0.022,  0.163,  0.305,  0.503]])

    Args:
        kernel_estimator: Method used to
            calculate the hat matrix (default =
            :class:`~skfda.misc.NadarayaWatsonHatMatrix`)
        weights: weight coefficients for each point.
        output_points: The output points. If omitted, the
            input points are used.

    So far only non parametric methods are implemented because we are only
    relying on a discrete representation of functional data.

    """

    def __init__(
        self,
        kernel_estimator: Optional[HatMatrix] = None,
        *,
        weights: Optional[NDArrayFloat] = None,
        output_points: Optional[GridPointsLike] = None,
        metric: Metric[NDArrayFloat] = l2_distance,
    ):
        self.kernel_estimator = kernel_estimator
        self.weights = weights
        self.output_points = output_points
        self.metric = metric
        self._cv = False  # For testing purposes only

    def _hat_matrix(
        self,
        input_points: GridPointsLike,
        output_points: GridPointsLike,
    ) -> NDArrayFloat:

        input_points = _cartesian_product(_to_grid_points(input_points))
        output_points = _cartesian_product(_to_grid_points(output_points))

        if self.kernel_estimator is None:
            self.kernel_estimator = NadarayaWatsonHatMatrix()

        delta_x = PairwiseMetric(self.metric)(output_points, input_points)

        return self.kernel_estimator(
            delta_x=delta_x,
            weights=self.weights,
            X_train=np.array(input_points),
            X=np.array(output_points),
            _cv=self._cv,
        )



"""Tests for FDataGrid's interpolation."""
import unittest

import numpy as np

from skfda import FDataGrid
from skfda.representation.interpolation import SplineInterpolation


class TestEvaluationSplineUnivariate(unittest.TestCase):
    """Test the evaluation of univariate spline interpolation."""

    def setUp(self) -> None:
        """
        Define the data.

        The data matrix consists on functions (x**2, (9-x)**2).

        """
        self.data_matrix_1_1 = [
            np.arange(10)**2,
            np.arange(start=9, stop=-1, step=-1)**2,
        ]

        self.grid_points = np.arange(10)

    def test_evaluation_linear_simple(self) -> None:
        """Test basic usage of linear evaluation."""
        f = FDataGrid(self.data_matrix_1_1, grid_points=self.grid_points)

        # Test interpolation in nodes
        np.testing.assert_allclose(
            f(self.grid_points)[..., 0],
            self.data_matrix_1_1,
        )

        # Test evaluation in a list of times
        np.testing.assert_allclose(
            f([0.5, 1.5, 2.5]),
            np.array([
                [[0.5], [2.5], [6.5]],
                [[72.5], [56.5], [42.5]],
            ]),
        )

    def test_evaluation_linear_point(self) -> None:
        """Test the evaluation of a single point."""
        f = FDataGrid(self.data_matrix_1_1, grid_points=self.grid_points)

        # Test a single point
        np.testing.assert_allclose(
            f(5.3),
            np.array([[[28.3]], [[13.9]]]),
        )
        np.testing.assert_allclose(
            f([3]), np.array([[[9]], [[36]]]),
        )
        np.testing.assert_allclose(
            f((2,)), np.array([[[4]], [[49]]]),
        )

    def test_evaluation_linear_grid(self) -> None:
        """Test grid evaluation. With domain dimension = 1."""
        f = FDataGrid(self.data_matrix_1_1, grid_points=self.grid_points)

        # Test interpolation in nodes
        np.testing.assert_allclose(
            f(self.grid_points)[..., 0],
            self.data_matrix_1_1,
        )

        res = np.array([[[0.5], [2.5], [6.5]], [[72.5], [56.5], [42.5]]])
        t = [0.5, 1.5, 2.5]

        # Test evaluation in a list of times
        np.testing.assert_allclose(f(t, grid=True), res)
        np.testing.assert_allclose(f((t,), grid=True), res)
        np.testing.assert_allclose(f([t], grid=True), res)
        # Single point with grid
        np.testing.assert_allclose(
            f(3, grid=True),
            np.array([[[9]], [[36]]]),
        )

        # Check erroneous axis
        with self.assertRaises(ValueError):
            f((t, t), grid=True)

    def test_evaluation_linear_unaligned(self) -> None:
        """Test unaligned evaluation."""
        f = FDataGrid(self.data_matrix_1_1, grid_points=self.grid_points)

        np.testing.assert_allclose(
            f([[1], [8]], aligned=False),
            np.array([[[1]], [[1]]]),
        )

        t = np.linspace(4, 6, 4)
        np.testing.assert_allclose(
            f([t, 9 - t], aligned=False),
            np.array([
                [[16], [22], [28.67], [36]],
                [[16], [22], [28.67], [36]],
            ]),
            rtol=1e-3,
        )

        # Same length than nsample
        t = np.linspace(4, 6, 2)
        np.testing.assert_allclose(
            f([t, 9 - t], aligned=False),
            np.array([[[16], [36]], [[16], [36]]]),
        )

    def test_evaluation_cubic_simple(self) -> None:
        """Test basic usage of cubic evaluation."""
        f = FDataGrid(
            self.data_matrix_1_1,
            grid_points=self.grid_points,
            interpolation=SplineInterpolation(3),
        )

        # Test interpolation in nodes
        np.testing.assert_allclose(
            f(self.grid_points)[..., 0],
            self.data_matrix_1_1,
            atol=1e-8,
        )

        # Test evaluation in a list of times
        np.testing.assert_allclose(
            f([0.5, 1.5, 2.5]),
            np.array([
                [[0.25], [2.25], [6.25]],
                [[72.25], [56.25], [42.25]],
            ]),
        )

    def test_evaluation_cubic_point(self) -> None:
        """Test the evaluation of a single point."""
        f = FDataGrid(
            self.data_matrix_1_1,
            grid_points=self.grid_points,
            interpolation=SplineInterpolation(3),
        )

        # Test a single point
        np.testing.assert_allclose(
            f(5.3),
            np.array([[[28.09]], [[13.69]]]),
        )

        np.testing.assert_allclose(
            f([3]),
            np.array([[[9]], [[36]]]),
        )
        np.testing.assert_allclose(
            f((2,)),
            np.array([[[4]], [[49]]]),
        )

    def test_evaluation_cubic_grid(self) -> None:
        """Test cubic grid evaluation."""
        f = FDataGrid(
            self.data_matrix_1_1,
            grid_points=self.grid_points,
            interpolation=SplineInterpolation(3),
        )

        t = [0.5, 1.5, 2.5]
        res = np.array([
            [[0.25], [2.25], [6.25]],
            [[72.25], [56.25], [42.25]],
        ])

        # Test evaluation in a list of times
        np.testing.assert_allclose(f(t, grid=True), res)
        np.testing.assert_allclose(f((t,), grid=True), res)
        np.testing.assert_allclose(f([t], grid=True), res)

        # Single point with grid
        np.testing.assert_allclose(
            f(3, grid=True),
            np.array([[[9]], [[36]]]),
        )

        # Check erroneous axis
        with self.assertRaises(ValueError):
            f((t, t), grid=True)

    def test_evaluation_cubic_unaligned(self) -> None:
        """Test cubic unaligned evaluation."""
        f = FDataGrid(
            self.data_matrix_1_1,
            grid_points=self.grid_points,
            interpolation=SplineInterpolation(3),
        )

        np.testing.assert_allclose(
            f([[1], [8]], aligned=False),
            np.array([[[1]], [[1]]]),
        )

        t = np.linspace(4, 6, 4)
        np.testing.assert_allclose(
            f([t, 9 - t], aligned=False),
            np.array([
                [[16], [21.78], [28.44], [36]],
                [[16], [21.78], [28.44], [36]],
            ]),
            rtol=1e-3,
        )

        # Same length than nsample
        t = np.linspace(4, 6, 2)
        np.testing.assert_allclose(
            f([t, 9 - t], aligned=False),
            np.array([
                [[16], [36]], [[16], [36]],
            ]),
        )

    def test_evaluation_nodes(self) -> None:
        """Test interpolation in nodes for all dimensions."""
        for degree in range(1, 6, 2):
            interpolation = SplineInterpolation(degree)

            f = FDataGrid(
                self.data_matrix_1_1,
                grid_points=self.grid_points,
                interpolation=interpolation,
            )

            # Test interpolation in nodes
            np.testing.assert_allclose(
                f(self.grid_points)[..., 0],
                self.data_matrix_1_1,
                atol=1e-8,
            )

    def test_error_degree(self) -> None:
        """Check unsupported spline degrees."""
        with self.assertRaises(ValueError):
            interpolation = SplineInterpolation(7)
            f = FDataGrid(
                self.data_matrix_1_1,
                grid_points=self.grid_points,
                interpolation=interpolation,
            )
            f(1)

        with self.assertRaises(ValueError):
            interpolation = SplineInterpolation(0)
            f = FDataGrid(
                self.data_matrix_1_1,
                grid_points=self.grid_points,
                interpolation=interpolation,
            )
            f(1)


class TestEvaluationSplineArbitraryImage(unittest.TestCase):
    """Test spline for arbitary image dimension."""

    def setUp(self) -> None:
        """
        Define the data.

        The data matrix consists on functions (x**2, (9-x)**2).

        """
        self.t = np.arange(10)

        data_1 = np.array([
            np.arange(10)**2,
            np.arange(start=9, stop=-1, step=-1)**2,
        ])
        data_2 = np.sin(np.pi / 81 * data_1)

        self.data_matrix_1_n = np.dstack((data_1, data_2))

        self.interpolation = SplineInterpolation(interpolation_order=2)

    def test_evaluation_simple(self) -> None:
        """Test basic usage of evaluation."""
        f = FDataGrid(
            self.data_matrix_1_n,
            grid_points=np.arange(10),
            interpolation=self.interpolation,
        )

        # Test interpolation in nodes
        np.testing.assert_allclose(
            f(self.t),
            self.data_matrix_1_n,
            atol=1e-8,
        )
        # Test evaluation in a list of times
        np.testing.assert_allclose(
            f([1.5, 2.5, 3.5]),
            np.array([
                [
                    [2.25, 0.087212],
                    [6.25, 0.240202],
                    [12.25, 0.45773],
                ],
                [
                    [56.25, 0.816142],
                    [42.25, 0.997589],
                    [30.25, 0.922146],
                ],
            ]),
            rtol=1e-5,
        )

    def test_evaluation_point(self) -> None:
        """Test the evaluation of a single point."""
        f = FDataGrid(
            self.data_matrix_1_n,
            grid_points=np.arange(10),
            interpolation=self.interpolation,
        )

        # Test a single point
        np.testing.assert_allclose(
            f(5.3),
            np.array([
                [[28.09, 0.885526]],
                [[13.69, 0.50697]],
            ]),
            rtol=1e-6,
        )

    def test_evaluation_grid(self) -> None:
        """Test grid evaluation."""
        f = FDataGrid(
            self.data_matrix_1_n,
            grid_points=np.arange(10),
            interpolation=SplineInterpolation(2),
        )

        t = [1.5, 2.5, 3.5]
        res = np.array([
            [
                [2.25, 0.08721158],
                [6.25, 0.24020233],
                [12.25, 0.4577302],
            ],
            [
                [56.25, 0.81614206],
                [42.25, 0.99758925],
                [30.25, 0.92214607],
            ],
        ])

        # Test evaluation in a list of times
        np.testing.assert_allclose(f(t, grid=True), res)
        np.testing.assert_allclose(f((t,), grid=True), res)
        np.testing.assert_allclose(f([t], grid=True), res)

        # Check erroneous axis
        with self.assertRaises(ValueError):
            f((t, t), grid=True)

    def test_evaluation_unaligned(self) -> None:
        """Test unaligned evaluation."""
        f = FDataGrid(
            self.data_matrix_1_n,
            grid_points=self.t,
            interpolation=self.interpolation,
        )

        np.testing.assert_allclose(
            f([[1], [4]], aligned=False)[0],
            f(1)[0],
        )
        np.testing.assert_allclose(
            f([[1], [4]], aligned=False)[1],
            f(4)[1],
        )

    def test_evaluation_nodes(self) -> None:
        """Test interpolation in nodes for all dimensions."""
        for degree in range(1, 6, 2):
            interpolation = SplineInterpolation(degree)

            f = FDataGrid(
                self.data_matrix_1_n,
                grid_points=np.arange(10),
                interpolation=interpolation,
            )

            # Test interpolation in nodes
            np.testing.assert_allclose(
                f(np.arange(10)),
                self.data_matrix_1_n,
                atol=1e-8,
            )


@np.vectorize
def _coordinate_function(
    *args: float,
) -> np.typing.NDArray[np.float_]:
    _, *domain_indexes, _ = args
    return np.sum(domain_indexes)  # type: ignore[no-any-return]


class TestEvaluationSplineArbitraryDim(unittest.TestCase):
    """Test spline for arbitrary domain and image dimensions."""

    def test_evaluation_middle_linear(self) -> None:
        """Test linear interpolation in the middle point of a grid square."""
        dim_codomain = 4
        n_samples = 2

        for dim_domain in range(1, 6):
            grid_points = [np.array([0, 1]) for _ in range(dim_domain)]
            data_matrix = np.fromfunction(
                function=_coordinate_function,
                shape=(n_samples,) + (2,) * dim_domain + (dim_codomain,),
            )

            f = FDataGrid(data_matrix, grid_points=grid_points)

            evaluation = f([
                [0] * dim_domain,
                [0.5] * dim_domain,
                [1] * dim_domain,
            ])

            self.assertEqual(evaluation.shape, (n_samples, 3, dim_codomain))

            for i in range(n_samples):
                for j in range(dim_codomain):
                    np.testing.assert_allclose(
                        evaluation[i, ..., j],
                        [0, dim_domain * 0.5, dim_domain],
                    )


if __name__ == '__main__':
    unittest.main()



"""Metrics, norms and related utilities."""

from typing import TYPE_CHECKING

import lazy_loader as lazy

__getattr__, __dir__, __all__ = lazy.attach(
    __name__,
    submod_attrs={
        "_angular": ["angular_distance"],
        "_fisher_rao": [
            "_fisher_rao_warping_distance",
            "fisher_rao_amplitude_distance",
            "fisher_rao_distance",
            "fisher_rao_phase_distance",
        ],
        "_lp_distances": [
            "LpDistance",
            "l1_distance",
            "l2_distance",
            "linf_distance",
            "lp_distance",
        ],
        "_lp_norms": [
            "LpNorm",
            "l1_norm",
            "l2_norm",
            "linf_norm",
            "lp_norm",
        ],
        "_mahalanobis": ["MahalanobisDistance"],
        "_parse": ["PRECOMPUTED"],
        "_utils": [
            "NormInducedMetric",
            "PairwiseMetric",
            "TransformationMetric",
            "pairwise_metric_optimization",
        ],
    },
)

if TYPE_CHECKING:
    from ._angular import angular_distance as angular_distance
    from ._fisher_rao import (
        _fisher_rao_warping_distance as _fisher_rao_warping_distance,
        fisher_rao_amplitude_distance as fisher_rao_amplitude_distance,
        fisher_rao_distance as fisher_rao_distance,
        fisher_rao_phase_distance as fisher_rao_phase_distance,
    )
    from ._lp_distances import (
        LpDistance as LpDistance,
        l1_distance as l1_distance,
        l2_distance as l2_distance,
        linf_distance as linf_distance,
        lp_distance as lp_distance,
    )
    from ._lp_norms import (
        LpNorm as LpNorm,
        l1_norm as l1_norm,
        l2_norm as l2_norm,
        linf_norm as linf_norm,
        lp_norm as lp_norm,
    )
    from ._mahalanobis import MahalanobisDistance as MahalanobisDistance
    from ._parse import PRECOMPUTED as PRECOMPUTED
    from ._utils import (
        NormInducedMetric as NormInducedMetric,
        PairwiseMetric as PairwiseMetric,
        TransformationMetric as TransformationMetric,
        pairwise_metric_optimization as pairwise_metric_optimization,
    )



"""Basis representation."""
from typing import TYPE_CHECKING

import lazy_loader as lazy

__getattr__, __dir__, __all__ = lazy.attach(
    __name__,
    submod_attrs={
        '_basis': ["Basis"],
        "_bspline_basis": ["BSplineBasis", "BSpline"],
        "_constant_basis": ["ConstantBasis", "Constant"],
        '_custom_basis': ["CustomBasis"],
        "_fdatabasis": ["FDataBasis", "FDataBasisDType"],
        "_finite_element_basis": ["FiniteElementBasis", "FiniteElement"],
        "_fourier_basis": ["FourierBasis", "Fourier"],
        "_grid_basis": ["_GridBasis"],
        "_monomial_basis": ["MonomialBasis", "Monomial"],
        "_tensor_basis": ["TensorBasis", "Tensor"],
        "_vector_basis": ["VectorValuedBasis", "VectorValued"],
    },
)

if TYPE_CHECKING:
    from ._basis import Basis as Basis
    from ._bspline_basis import (
        BSpline as BSpline,
        BSplineBasis as BSplineBasis,
    )
    from ._constant_basis import (
        Constant as Constant,
        ConstantBasis as ConstantBasis,
    )
    from ._custom_basis import CustomBasis as CustomBasis
    from ._fdatabasis import (
        FDataBasis as FDataBasis,
        FDataBasisDType as FDataBasisDType,
    )
    from ._finite_element_basis import (
        FiniteElement as FiniteElement,
        FiniteElementBasis as FiniteElementBasis,
    )
    from ._fourier_basis import (
        Fourier as Fourier,
        FourierBasis as FourierBasis,
    )
    from ._grid_basis import _GridBasis as _GridBasis
    from ._monomial_basis import (
        Monomial as Monomial,
        MonomialBasis as MonomialBasis,
    )
    from ._tensor_basis import Tensor as Tensor, TensorBasis as TensorBasis
    from ._vector_basis import (
        VectorValued as VectorValued,
        VectorValuedBasis as VectorValuedBasis,
    )



"""Tests of compatibility between numpy ufuncs and FDataGrid."""

import unittest
from typing import Any, Callable, Protocol, TypeVar

import numpy as np
import pytest

from skfda import FDataGrid
from skfda.representation.basis import FourierBasis

T = TypeVar("T", "np.typing.NDArray[Any]", FDataGrid)


class _MonaryUfuncProtocol(Protocol):

    def __call__(self, __arg: T) -> T:  # noqa: WPS112
        pass


@pytest.fixture(params=[
    np.sqrt,
    np.absolute,
    np.round,
    np.exp,
    np.log,
    np.log10,
    np.log2,
])
def monary(request: Any) -> Any:
    """
    Fixture providing the monary function to validate.

    Not all of them are ufuncs.

    """
    return request.param


def test_monary_ufuncs(monary: _MonaryUfuncProtocol) -> None:
    """Test that unary ufuncs can be applied to FDataGrid."""
    data_matrix = np.arange(15).reshape(3, 5) + 1

    fd = FDataGrid(data_matrix)

    fd_monary = monary(fd)

    fd_monary_build = FDataGrid(monary(data_matrix))

    assert fd_monary.equals(fd_monary_build)


def test_binary_ufunc() -> None:
    """Test that binary ufuncs can be applied to FDataGrid."""
    data_matrix = np.arange(15).reshape(3, 5)
    data_matrix2 = 2 * np.arange(15).reshape(3, 5)

    fd = FDataGrid(data_matrix)
    fd2 = FDataGrid(data_matrix2)

    fd_mul = np.multiply(fd, fd2)

    fd_mul_build = FDataGrid(data_matrix * data_matrix2)

    assert fd_mul.equals(fd_mul_build)


def test_out_ufunc(monary: Callable[..., Any]) -> None:
    """Test that the out parameter of ufuncs work for FDataGrid."""
    data_matrix = np.arange(15).reshape(3, 5) + 1
    data_matrix_copy = np.copy(data_matrix)

    fd = FDataGrid(data_matrix)

    monary(fd, out=fd)

    fd_monary_build = FDataGrid(monary(data_matrix_copy))

    assert fd.equals(fd_monary_build)


class TestOperators(unittest.TestCase):
    """Tests for operators."""

    def test_commutativity(self) -> None:
        """Test that operations with numpy arrays commute."""
        X = FDataGrid([[1, 2, 3], [4, 5, 6]])
        arr = np.array([1, 2])

        self.assertTrue((arr + X).equals((X + arr)))

    def test_commutativity_basis(self) -> None:
        """Test that operations with numpy arrays for basis commute."""
        X = FDataGrid([[1, 2, 3], [4, 5, 6]])
        arr = np.array([1, 2])
        basis = FourierBasis(n_basis=5)

        X_basis = X.to_basis(basis)

        self.assertTrue((arr * X_basis).equals((X_basis * arr)))



"""Statistics."""

from typing import TYPE_CHECKING

import lazy_loader as lazy

__getattr__, __dir__, __all__ = lazy.attach(
    __name__,
    submodules=["covariance"],
    submod_attrs={
        "_fisher_rao": [
            "_fisher_rao_warping_mean",
            "fisher_rao_karcher_mean",
        ],
        "_stats": [
            "cov",
            "depth_based_median",
            "geometric_median",
            "gmean",
            "mean",
            "modified_epigraph_index",
            "trim_mean",
            "var",
        ],
    },
)

if TYPE_CHECKING:
    from ._fisher_rao import (
        _fisher_rao_warping_mean as _fisher_rao_warping_mean,
        fisher_rao_karcher_mean as fisher_rao_karcher_mean,
    )
    from ._stats import (
        cov as cov,
        depth_based_median as depth_based_median,
        geometric_median as geometric_median,
        gmean as gmean,
        mean as mean,
        modified_epigraph_index as modified_epigraph_index,
        trim_mean as trim_mean,
        var as var,
    )



"""Test to check the extrapolation module."""

import unittest

import numpy as np

from skfda import FDataBasis, FDataGrid
from skfda.datasets import make_sinusoidal_process
from skfda.representation.basis import FourierBasis
from skfda.representation.extrapolation import (
    BoundaryExtrapolation,
    ExceptionExtrapolation,
    FillExtrapolation,
    PeriodicExtrapolation,
)


class TestExtrapolation(unittest.TestCase):
    """Extrapolation tests."""

    def setUp(self) -> None:
        """Create example data."""
        self.grid = make_sinusoidal_process(n_samples=2, random_state=0)
        self.basis = self.grid.to_basis(FourierBasis())
        self.dummy_data = [[1, 2, 3], [2, 3, 4]]

    def test_constructor_fdatabasis_setting(self) -> None:
        """Check argument normalization in constructor for FDataBasis."""
        coeff = self.dummy_data
        basis = FourierBasis(n_basis=3)

        a = FDataBasis(basis, coeff)
        self.assertEqual(a.extrapolation, None)

        a = FDataBasis(basis, coeff, extrapolation="periodic")
        self.assertEqual(a.extrapolation, PeriodicExtrapolation())

        self.assertNotEqual(a.extrapolation, BoundaryExtrapolation())

        a = FDataBasis(basis, coeff, extrapolation=BoundaryExtrapolation())
        self.assertEqual(a.extrapolation, BoundaryExtrapolation())

        a = FDataBasis(basis, coeff, extrapolation="exception")
        self.assertEqual(a.extrapolation, ExceptionExtrapolation())

        a = FDataBasis(basis, coeff, extrapolation=FillExtrapolation(0))
        self.assertEqual(a.extrapolation, FillExtrapolation(0))
        self.assertNotEqual(a.extrapolation, FillExtrapolation(1))

    def test_constructor_fdatagrid_setting(self) -> None:
        """Check argument normalization in constructor for FDataGrid."""
        data = self.dummy_data

        a = FDataGrid(data)
        self.assertEqual(a.extrapolation, None)

        a = FDataGrid(data, extrapolation="periodic")
        self.assertEqual(a.extrapolation, PeriodicExtrapolation())

        a = FDataGrid(data, extrapolation=BoundaryExtrapolation())
        self.assertEqual(a.extrapolation, BoundaryExtrapolation())

        self.assertNotEqual(a.extrapolation, ExceptionExtrapolation())

        a = FDataGrid(data, extrapolation="exception")
        self.assertEqual(a.extrapolation, ExceptionExtrapolation())

        a = FDataGrid(data, extrapolation=FillExtrapolation(0))
        self.assertEqual(a.extrapolation, FillExtrapolation(0))
        self.assertNotEqual(a.extrapolation, FillExtrapolation(1))

    def test_setting(self) -> None:
        """Check argument in setter."""
        data = self.dummy_data
        a = FDataGrid(data)

        a.extrapolation = PeriodicExtrapolation()
        self.assertEqual(a.extrapolation, PeriodicExtrapolation())

        a.extrapolation = "bounds"  # type: ignore[assignment]
        self.assertEqual(a.extrapolation, BoundaryExtrapolation())

        a.extrapolation = ExceptionExtrapolation()
        self.assertEqual(a.extrapolation, ExceptionExtrapolation())

        a.extrapolation = "zeros"  # type: ignore[assignment]
        self.assertEqual(a.extrapolation, FillExtrapolation(0))

        self.assertNotEqual(a.extrapolation, FillExtrapolation(1))

    def test_periodic(self) -> None:
        """Test periodic extrapolation."""
        self.grid.extrapolation = PeriodicExtrapolation()
        data = self.grid([-0.5, 0, 1.5])

        np.testing.assert_allclose(
            data[..., 0],
            [
                [-0.723516, 0.976450, -0.723516],
                [-1.085999, 0.759385, -1.085999],
            ],
            rtol=1e-6,
        )

        self.basis.extrapolation = "periodic"  # type: ignore[assignment]
        data = self.basis([-0.5, 0, 1.5])

        np.testing.assert_allclose(
            data[..., 0],
            [
                [-0.690170, 0.691735, -0.690170],
                [-1.020821, 1.056383, -1.020821],
            ],
            rtol=1e-6,
        )

    def test_boundary(self) -> None:
        """Test boundary-copying extrapolation."""
        self.grid.extrapolation = "bounds"  # type: ignore[assignment]
        data = self.grid([-0.5, 0, 1.5])

        np.testing.assert_allclose(
            data[..., 0],
            [
                [0.976450, 0.976450, 0.796817],
                [0.759385, 0.759385, 1.125063],
            ],
            rtol=1e-6,
        )

        self.basis.extrapolation = "bounds"  # type: ignore[assignment]
        data = self.basis([-0.5, 0, 1.5])

        np.testing.assert_allclose(
            data[..., 0],
            [
                [0.691735, 0.691735, 0.691735],
                [1.056383, 1.056383, 1.056383],
            ],
            rtol=1e-6,
        )

    def test_exception(self) -> None:
        """Test no extrapolation (exception)."""
        self.grid.extrapolation = "exception"  # type: ignore[assignment]

        with np.testing.assert_raises(ValueError):
            self.grid([-0.5, 0, 1.5])

        self.basis.extrapolation = "exception"  # type: ignore[assignment]

        with np.testing.assert_raises(ValueError):
            self.basis([-0.5, 0, 1.5])

    def test_zeros(self) -> None:
        """Test zeros extrapolation."""
        self.grid.extrapolation = "zeros"  # type: ignore[assignment]
        data = self.grid([-0.5, 0, 1.5])

        np.testing.assert_allclose(
            data[..., 0],
            [
                [0, 0.976450, 0],
                [0, 0.759385, 0],
            ],
            rtol=1e-6,
        )

        self.basis.extrapolation = "zeros"  # type: ignore[assignment]
        data = self.basis([-0.5, 0, 1.5])

        np.testing.assert_allclose(
            data[..., 0],
            [
                [0, 0.691735, 0],
                [0, 1.056383, 0],
            ],
            rtol=1e-6,
        )

    def test_nan(self) -> None:
        """Test nan extrapolation."""
        self.grid.extrapolation = "nan"  # type: ignore[assignment]
        data = self.grid([-0.5, 0, 1.5])

        np.testing.assert_allclose(
            data[..., 0],
            [
                [np.nan, 0.976450, np.nan],
                [np.nan, 0.759385, np.nan],
            ],
            rtol=1e-6,
        )

        self.basis.extrapolation = "nan"  # type: ignore[assignment]
        data = self.basis([-0.5, 0, 1.5])

        np.testing.assert_allclose(
            data[..., 0],
            [
                [np.nan, 0.691735, np.nan],
                [np.nan, 1.056383, np.nan],
            ],
            rtol=1e-6,
        )


if __name__ == '__main__':
    unittest.main()



"""Depth."""

from typing import TYPE_CHECKING

import lazy_loader as lazy

__getattr__, __dir__, __all__ = lazy.attach(
    __name__,
    submodules=[
        "multivariate",
    ],
    submod_attrs={
        '_depth': [
            "BandDepth",
            "DistanceBasedDepth",
            "IntegratedDepth",
            "ModifiedBandDepth",
        ],
        "multivariate": [
            "Depth",
            "Outlyingness",
            "OutlyingnessBasedDepth",
        ],
    },
)

if TYPE_CHECKING:
    from ._depth import (
        BandDepth as BandDepth,
        DistanceBasedDepth as DistanceBasedDepth,
        IntegratedDepth as IntegratedDepth,
        ModifiedBandDepth as ModifiedBandDepth,
    )
    from .multivariate import (
        Depth as Depth,
        Outlyingness as Outlyingness,
        OutlyingnessBasedDepth as OutlyingnessBasedDepth,
    )



"""Test for scoring module."""
import math
import unittest
from typing import Any, Optional, Sequence, Tuple

import numpy as np
import sklearn.metrics

from skfda import FDataBasis, FDataGrid
from skfda.misc.scoring import (
    ScoreFunction,
    explained_variance_score,
    mean_absolute_error,
    mean_absolute_percentage_error,
    mean_squared_error,
    mean_squared_log_error,
    r2_score,
)
from skfda.representation.basis import (
    BSplineBasis,
    FourierBasis,
    MonomialBasis,
)
from skfda.typing._numpy import NDArrayFloat

score_functions: Sequence[ScoreFunction] = (
    explained_variance_score,
    mean_absolute_error,
    mean_absolute_percentage_error,
    mean_squared_error,
    mean_squared_log_error,
    r2_score,
)


def _create_data_basis() -> Tuple[FDataBasis, FDataBasis]:
    coef_true = [[1, 2, 3], [4, 5, 6]]
    coef_pred = [[1, 2, 3], [4, 6, 5]]

    # y_true: 1) 1 + 2x + 3x^2
    #         2) 4 + 5x + 6x^2
    y_true = FDataBasis(
        basis=MonomialBasis(domain_range=((0, 3),), n_basis=3),
        coefficients=coef_true,
    )

    # y_pred: 1) 1 + 2x + 3x^2
    #         2) 4 + 6x + 5x^2
    y_pred = FDataBasis(
        basis=MonomialBasis(domain_range=((0, 3),), n_basis=3),
        coefficients=coef_pred,
    )

    # y_true - y_pred: 1) 0
    #                  2) -x + x^2
    return y_true, y_pred


def _create_data_grid() -> Tuple[FDataGrid, FDataGrid]:

    y_true, y_pred = _create_data_basis()
    grid = np.linspace(*y_true.domain_range[0], 100)

    return y_true.to_grid(grid), y_pred.to_grid(grid)


class TestScoreFunctionsGrid(unittest.TestCase):
    """Tests for score functions with FDataGrid representation."""

    def _test_generic_grid(
        self,
        function: ScoreFunction,
        weight: Optional[NDArrayFloat] = None,
        **kwargs: Any,
    ) -> None:
        y_true, y_pred = _create_data_grid()

        score = function(
            y_true,
            y_pred,
            multioutput='raw_values',
            sample_weight=weight,
            **kwargs,
        )

        score_sklearn = function(
            y_true.data_matrix.squeeze(),
            y_pred.data_matrix.squeeze(),
            multioutput='raw_values',
            sample_weight=weight,
            **kwargs,
        )

        np.testing.assert_allclose(
            score.data_matrix.squeeze(),
            score_sklearn,
        )

    def test_all(self) -> None:
        """Test all score functions."""
        for score_function in score_functions:
            with self.subTest(function=score_function):

                self._test_generic_grid(score_function)

                try:
                    self._test_generic_grid(
                        score_function,
                        squared=False,
                    )
                except TypeError:
                    pass

                self._test_generic_grid(
                    score_function,
                    weight=np.array([3, 1]),
                )


class TestScoreFunctionGridBasis(unittest.TestCase):
    """Compare the results obtained for FDataGrid and FDataBasis."""

    def _test_grid_basis_generic(
        self,
        score_function: ScoreFunction,
        weight: Optional[NDArrayFloat] = None,
        **kwargs: Any,
    ) -> None:
        y_true_grid, y_pred_grid = _create_data_grid()

        y_true_basis = y_true_grid.to_basis(basis=BSplineBasis(n_basis=10))
        y_pred_basis = y_pred_grid.to_basis(basis=BSplineBasis(n_basis=10))

        # The results should be close but not equal as the two representations
        # do not give same functions.
        precision = 2

        score_grid = score_function(
            y_true_grid,
            y_pred_grid,
            sample_weight=weight,
            **kwargs,
        )
        score_basis = score_function(
            y_true_basis,
            y_pred_basis,
            sample_weight=weight,
            **kwargs,
        )

        self.assertAlmostEqual(score_basis, score_grid, places=precision)

    def test_all(self) -> None:
        """Test all score functions."""
        for score_function in score_functions:
            with self.subTest(function=score_function):

                self._test_grid_basis_generic(score_function)

                try:
                    self._test_grid_basis_generic(
                        score_function,
                        squared=False,
                    )
                except TypeError:
                    pass

                self._test_grid_basis_generic(
                    score_function,
                    weight=np.array([3, 1]),
                )


class TestScoreFunctionsBasis(unittest.TestCase):
    """Tests for score functions with FDataBasis representation."""

    def test_explained_variance_basis(self) -> None:
        """Test Explain Variance Score for FDataBasis."""
        y_true, y_pred = _create_data_basis()

        ev = explained_variance_score(y_true, y_pred)

        # integrate 1 - num/den
        # where     num = (1/2x -1/2x^2)^2
        # and       den = (1.5 + 1.5x + 1.5x^2)^2
        self.assertAlmostEqual(ev, 0.992968, places=6)

    def test_mean_absolut_error_basis(self) -> None:
        """Test Mean Absolute Error for FDataBasis."""
        y_true, y_pred = _create_data_basis()

        mae = mean_absolute_error(y_true, y_pred)

        # integrate 1/2 * | -x + x^2|
        self.assertAlmostEqual(mae, 0.8055555555)

    def test_mean_absolute_percentage_error_basis(self) -> None:
        """Test Mean Absolute Percentage Error for FDataBasis."""
        y_true, y_pred = _create_data_basis()

        mape = mean_absolute_percentage_error(y_true, y_pred)

        # integrate |1/2 * (-x  + x^2) / (4 + 5x + 6x^2)|
        self.assertAlmostEqual(mape, 0.0199192187)

    def test_mean_squared_error_basis(self) -> None:
        """Test Mean Squared Error for FDataBasis."""
        y_true, y_pred = _create_data_basis()

        mse = mean_squared_error(y_true, y_pred)

        # integrate 1/2 * (-x + x^2)^2
        self.assertAlmostEqual(mse, 2.85)

    def test_mean_squared_log_error_basis(self) -> None:
        """Test Mean Squared Log Error for FDataBasis."""
        y_true, y_pred = _create_data_basis()

        msle = mean_squared_log_error(y_true, y_pred)

        # integrate 1/2*(log(1 + 4 + 5x + 6x^2) - log(1 + 4 + 6x + 5x^2))^2
        self.assertAlmostEqual(msle, 0.00107583)

    def test_r2_score_basis(self) -> None:
        """Test R2 Score for FDataBasis."""
        y_true, y_pred = _create_data_basis()

        r2 = r2_score(y_true, y_pred)

        # integrate 1 - num/den
        # where     num = 1/2*(-x + x^2)^2,
        # and       den = (1.5 + 1.5x + 1.5x^2)^2
        self.assertAlmostEqual(r2, 0.9859362)


class TestScoreZeroDenominator(unittest.TestCase):
    """Tests Score Functions with edge cases."""

    def test_zero_r2(self) -> None:
        """Test R2 Score when the denominator is zero."""
        # Case when both numerator and denominator is zero (in t = 1)
        basis_coef_true = [[0, 1], [-1, 2], [-2, 3]]
        basis_coef_pred = [[1, 0], [2, -1], [3, -2]]

        # y_true and y_pred are 2 sets of 3 straight lines
        # for all f, f(1) = 1
        y_true_basis = FDataBasis(
            basis=MonomialBasis(domain_range=((0, 2),), n_basis=2),
            coefficients=basis_coef_true,
        )

        y_pred_basis = FDataBasis(
            basis=MonomialBasis(domain_range=((0, 2),), n_basis=2),
            coefficients=basis_coef_pred,
        )

        grid_points = np.linspace(0, 2, 9)

        y_true_grid = y_true_basis.to_grid(grid_points=grid_points)
        y_pred_grid = y_pred_basis.to_grid(grid_points=grid_points)

        # 0/0 for FDataGrid
        np.testing.assert_almost_equal(
            r2_score(
                y_true_grid,
                y_pred_grid,
                multioutput='raw_values',
            ).data_matrix.squeeze(),
            sklearn.metrics.r2_score(
                y_true_grid.data_matrix.squeeze(),
                y_pred_grid.data_matrix.squeeze(),
                multioutput='raw_values',
            ),
        )

        # 0/0 for FDataBasis
        self.assertAlmostEqual(
            r2_score(y_true_basis, y_pred_basis),
            -16.5,
        )

        # Case when numerator is non-zero and denominator is zero (in t = 1)
        # for all f in y_true, f(1) = 1
        # for all f in y_pred, f(1) = 2
        basis_coef_pred = [[2, 0], [3, -1], [4, -2]]
        y_pred_basis = FDataBasis(
            basis=MonomialBasis(domain_range=((0, 2),), n_basis=2),
            coefficients=basis_coef_pred,
        )

        y_pred_grid = y_pred_basis.to_grid(grid_points=grid_points)

        # r/0 for FDataGrid (r != 0)
        np.testing.assert_almost_equal(
            r2_score(
                y_true_grid,
                y_pred_grid,
                multioutput='raw_values',
            )(1),
            [[[-math.inf]]],
        )

        # r/0 for FDataBasis (r != 0)
        self.assertAlmostEqual(
            r2_score(
                y_true_basis,
                y_pred_basis,
            ),
            -math.inf,
        )

    def test_zero_ev(self) -> None:
        """Test R2 Score when the denominator is zero."""
        basis_coef_true = [[0, 1], [-1, 2], [-2, 3]]
        basis_coef_pred = [[1, 0], [2, -1], [3, -2]]
        # Case when both numerator and denominator is zero (in t = 1)

        # y_true and y_pred are 2 sets of 3 straight lines
        # for all f, f(1) = 1
        # var(y_true(1)) = 0 and var(y_true(1) - y_pred(1)) = 0
        y_true_basis = FDataBasis(
            basis=MonomialBasis(domain_range=((0, 2),), n_basis=2),
            coefficients=basis_coef_true,
        )

        y_pred_basis = FDataBasis(
            basis=MonomialBasis(domain_range=((0, 2),), n_basis=2),
            coefficients=basis_coef_pred,
        )

        grid_points = np.linspace(0, 2, 9)

        y_true_grid = y_true_basis.to_grid(grid_points=grid_points)
        y_pred_grid = y_pred_basis.to_grid(grid_points=grid_points)

        # 0/0 for FDataGrid
        np.testing.assert_almost_equal(
            explained_variance_score(
                y_true_grid,
                y_pred_grid,
                multioutput='raw_values',
            ).data_matrix.flatten(),
            sklearn.metrics.explained_variance_score(
                y_true_grid.data_matrix.squeeze(),
                y_pred_grid.data_matrix.squeeze(),
                multioutput='raw_values',
            ),
        )

        # 0/0 for FDataBasis
        self.assertAlmostEqual(
            explained_variance_score(y_true_basis, y_pred_basis),
            -3,
        )

        # Case when numerator is non-zero and denominator is zero (in t = 1)
        basis_coef_pred = [[2, 0], [2, -1], [2, -2]]
        y_pred_basis = FDataBasis(
            basis=MonomialBasis(domain_range=((0, 2),), n_basis=2),
            coefficients=basis_coef_pred,
        )
        y_pred_grid = y_pred_basis.to_grid(grid_points=grid_points)

        # r/0 for FDataGrid (r != 0)
        np.testing.assert_almost_equal(
            explained_variance_score(
                y_true_grid,
                y_pred_grid,
                multioutput='raw_values',
            )(1),
            [[[-math.inf]]],
        )

        # r/0 for FDataBasis (r != 0)
        self.assertAlmostEqual(
            explained_variance_score(
                y_true_basis,
                y_pred_basis,
            ),
            -math.inf,
        )

    def test_zero_mape(self) -> None:
        """Test Mean Absolute Percentage Error when y_true can be zero."""
        basis_coef_true = [[3, 0, 0, 0, 0], [0, 0, 1, 0, 0]]
        basis_coef_pred = [[1, 0, 0, 0, 0], [0, 0, 1, 1, 0]]

        # Fourier basis defined in (0, 2) with 3 elements
        # The functions are
        # y_true(t) = 1) 3/sqrt(2)
        #             2) 1/sqrt(2) * cos(pi t)
        #
        # y_pred(t) = 1) 1/sqrt(2)
        #             2) 1/sqrt(2) * cos(pi t) + 1/sqrt(2) * sin(2 pi t)
        # The second function in y_true should be zero at t = 0.5 and t = 1.5

        y_true_basis = FDataBasis(
            basis=FourierBasis(domain_range=((0, 2),), n_basis=5),
            coefficients=basis_coef_true,
        )

        y_pred_basis = FDataBasis(
            basis=FourierBasis(domain_range=((0, 2),), n_basis=5),
            coefficients=basis_coef_pred,
        )

        with self.assertWarns(RuntimeWarning):
            mean_absolute_percentage_error(y_true_basis, y_pred_basis)

        grid_points = np.linspace(0, 2, 9)

        # The set of points in which the functions are evaluated
        # includes 0.5 and 1.5
        y_true_grid = y_true_basis.to_grid(grid_points=grid_points)
        y_pred_grid = y_pred_basis.to_grid(grid_points=grid_points)

        with self.assertWarns(RuntimeWarning):
            mean_absolute_percentage_error(y_true_grid, y_pred_grid)

    def test_negative_msle(self) -> None:
        """Test Mean Squared Log Error when there are negative data."""
        basis_coef_true = [[3, 0, 0], [0, 0, 1]]
        basis_coef_pred = [[1, 0, 0], [np.sqrt(2), 0, 1]]

        # Fourier basis defined in (0, 2) with 3 elements
        # The functions are
        # y_true(t) = 1) 3/sqrt(2)
        #             2) 1/sqrt(2) * cos(pi t)
        #
        # y_pred(t) = 1) 1/sqrt(2)
        #             2) 1 + 1/sqrt(2) * cos(pi t)
        # The second function in y_true should be negative
        # between t = 0.5 and t = 1.5
        # All functions in y_pred should be always positive

        y_true_basis = FDataBasis(
            basis=FourierBasis(domain_range=((0, 2),), n_basis=3),
            coefficients=basis_coef_true,
        )

        y_pred_basis = FDataBasis(
            basis=FourierBasis(domain_range=((0, 2),), n_basis=3),
            coefficients=basis_coef_pred,
        )

        self.assertRaises(
            ValueError,
            mean_squared_log_error,
            y_true_basis,
            y_pred_basis,
        )

        grid_points = np.linspace(0, 2, 9)

        y_true_grid = y_true_basis.to_grid(grid_points=grid_points)
        y_pred_grid = y_pred_basis.to_grid(grid_points=grid_points)

        self.assertRaises(
            ValueError,
            mean_squared_log_error,
            y_true_grid,
            y_pred_grid,
        )



"""
Depth Measures Module.

This module includes different methods to order functional data,
from the center (larger values) outwards(smaller ones).

"""
from __future__ import annotations

import itertools
from typing import TypeVar

import numpy as np

from ..._utils._sklearn_adapter import BaseEstimator
from ..._utils.ndfunction import average_function_value
from ...misc.metrics import l2_distance
from ...misc.metrics._utils import _fit_metric
from ...representation import FData, FDataGrid
from ...typing._metric import Metric
from ...typing._numpy import NDArrayFloat
from .multivariate import Depth, SimplicialDepth, _UnivariateFraimanMuniz

T = TypeVar("T", bound=FData)


class IntegratedDepth(Depth[FDataGrid]):
    r"""
    Functional depth as the integral of a multivariate depth.

    Args:
        multivariate_depth (Depth): Multivariate depth to integrate.
            By default it is the one used by Fraiman and Muniz, that is,

            .. math::
                D(x) = 1 - \left\lvert \frac{1}{2}- F(x)\right\rvert

    Examples:
        >>> import skfda
        >>>
        >>> data_matrix = [[1, 1, 2, 3, 2.5, 2],
        ...                [0.5, 0.5, 1, 2, 1.5, 1],
        ...                [-1, -1, -0.5, 1, 1, 0.5],
        ...                [-0.5, -0.5, -0.5, -1, -1, -1]]
        >>> grid_points = [0, 2, 4, 6, 8, 10]
        >>> fd = skfda.FDataGrid(data_matrix, grid_points)
        >>> depth = skfda.exploratory.depth.IntegratedDepth()
        >>> depth(fd)
        array([ 0.5  ,  0.75 ,  0.925,  0.875])

    References:
        Fraiman, R., & Muniz, G. (2001). Trimmed means for functional
        data. Test, 10(2), 419–440. https://doi.org/10.1007/BF02595706


    """

    def __init__(
        self,
        *,
        multivariate_depth: Depth[NDArrayFloat] | None = None,
    ) -> None:
        self.multivariate_depth = multivariate_depth

    def fit(  # noqa: D102
        self,
        X: FDataGrid,
        y: object = None,
    ) -> IntegratedDepth:

        self.multivariate_depth_: Depth[NDArrayFloat]

        if self.multivariate_depth is None:
            self.multivariate_depth_ = _UnivariateFraimanMuniz()
        else:
            self.multivariate_depth_ = self.multivariate_depth

        self._domain_range = X.domain_range
        self._grid_points = X.grid_points
        self.multivariate_depth_.fit(X.data_matrix)
        return self

    def transform(self, X: FDataGrid) -> NDArrayFloat:  # noqa: D102

        pointwise_depth = X.copy(
            data_matrix=self.multivariate_depth_.transform(X.data_matrix),
        )

        return average_function_value(pointwise_depth).ravel()

    @property
    def max(self) -> float:
        if self.multivariate_depth is None:
            return 1

        return self.multivariate_depth.max

    @property
    def min(self) -> float:
        if self.multivariate_depth is None:
            return 1 / 2

        return self.multivariate_depth.min


class ModifiedBandDepth(IntegratedDepth):
    """
    Implementation of Modified Band Depth for functional data.

    The band depth of each sample is obtained by computing the fraction of time
    its graph is contained in the bands determined by two sample curves.
    In the case the fdatagrid :term:`domain` dimension is 2, instead of curves,
    surfaces determine the bands. In larger dimensions, the hyperplanes
    determine the bands.

    Examples:
        >>> import skfda
        >>>
        >>> data_matrix = [[1, 1, 2, 3, 2.5, 2],
        ...                [0.5, 0.5, 1, 2, 1.5, 1],
        ...                [-1, -1, -0.5, 1, 1, 0.5],
        ...                [-0.5, -0.5, -0.5, -1, -1, -1]]
        >>> grid_points = [0, 2, 4, 6, 8, 10]
        >>> fd = skfda.FDataGrid(data_matrix, grid_points)
        >>> depth = skfda.exploratory.depth.ModifiedBandDepth()
        >>> values = depth(fd)
        >>> values.round(2)
        array([ 0.5 ,  0.83,  0.73,  0.67])

    References:
        López-Pintado, S., & Romo, J. (2009). On the Concept of
        Depth for Functional Data. Journal of the American Statistical
        Association, 104(486), 718–734.
        https://doi.org/10.1198/jasa.2009.0108
    """

    def __init__(self) -> None:
        super().__init__(multivariate_depth=SimplicialDepth())


class BandDepth(Depth[FDataGrid]):
    """
    Implementation of Band Depth for functional data.

    The band depth of each sample is obtained by computing the fraction of the
    bands determined by two sample curves containing the whole graph of the
    first one. In the case the fdatagrid :term:`domain` dimension is 2, instead
    of curves, surfaces determine the bands. In larger dimensions, the
    hyperplanes determine the bands.

    Examples:
        >>> import skfda
        >>>
        >>> data_matrix = [[1, 1, 2, 3, 2.5, 2],
        ...                [0.5, 0.5, 1, 2, 1.5, 1],
        ...                [-1, -1, -0.5, 1, 1, 0.5],
        ...                [-0.5, -0.5, -0.5, -1, -1, -1]]
        >>> grid_points = [0, 2, 4, 6, 8, 10]
        >>> fd = skfda.FDataGrid(data_matrix, grid_points)
        >>> depth = skfda.exploratory.depth.BandDepth()
        >>> depth(fd)
        array([ 0.5       ,  0.83333333,  0.5       ,  0.5       ])

    References:
        López-Pintado, S., & Romo, J. (2009). On the Concept of
        Depth for Functional Data. Journal of the American Statistical
        Association, 104(486), 718–734.
        https://doi.org/10.1198/jasa.2009.0108

    """

    def fit(self, X: FDataGrid, y: object = None) -> BandDepth:  # noqa: D102

        if X.dim_codomain != 1:
            raise NotImplementedError(
                "Band depth not implemented for vector valued functions",
            )

        self._distribution = X
        return self

    def transform(self, X: FDataGrid) -> NDArrayFloat:  # noqa: D102

        num_in = np.zeros(shape=len(X), dtype=X.data_matrix.dtype)
        n_total = 0

        for f1, f2 in itertools.combinations(self._distribution, 2):
            between_range_1 = (
                (f1.data_matrix <= X.data_matrix)
                & (X.data_matrix <= f2.data_matrix)
            )

            between_range_2 = (
                (f2.data_matrix <= X.data_matrix)
                & (X.data_matrix <= f1.data_matrix)
            )

            between_range = between_range_1 | between_range_2

            num_in += np.all(
                between_range,
                axis=tuple(range(1, X.data_matrix.ndim)),
            )
            n_total += 1

        return num_in / n_total


class DistanceBasedDepth(Depth[FDataGrid], BaseEstimator):
    r"""
    Functional depth based on a metric.

    Parameters:
        metric:
            The metric to use as M in the following depth calculation

            .. math::
                D(x) = [1 + M(x, \mu)]^{-1}.

            as explained in :footcite:`serfling+zuo_2000_general`.

    Examples:
        >>> import skfda
        >>> from skfda.exploratory.depth import DistanceBasedDepth
        >>> from skfda.misc.metrics import MahalanobisDistance
        >>> data_matrix = [[1, 1, 2, 3, 2.5, 2],
        ...                [0.5, 0.5, 1, 2, 1.5, 1],
        ...                [-1, -1, -0.5, 1, 1, 0.5],
        ...                [-0.5, -0.5, -0.5, -1, -1, -1]]
        >>> grid_points = [0, 2, 4, 6, 8, 10]
        >>> fd = skfda.FDataGrid(data_matrix, grid_points)
        >>> depth = DistanceBasedDepth(MahalanobisDistance(2))
        >>> depth(fd)
        array([ 0.41897777,  0.8058132 ,  0.31097392,  0.31723619])

    References:
        .. footbibliography::
    """

    def __init__(
        self,
        metric: Metric[T] = l2_distance,
    ) -> None:
        self.metric = metric

    def fit(  # noqa: D102
        self,
        X: T,
        y: object = None,
    ) -> DistanceBasedDepth:
        """Fit the model using X as training data.

        Args:
            X: FDataGrid with the training data or array matrix with shape
                (n_samples, n_samples) if metric='precomputed'.
            y: Ignored.

        Returns:
            self
        """
        _fit_metric(self.metric, X)

        self.mean_ = X.mean()

        return self

    def transform(self, X: T) -> NDArrayFloat:  # noqa: D102
        """Compute the depth of given observations.

        Args:
            X: FDataGrid with the observations to use in the calculation.

        Returns:
            Array with the depths.
        """
        return 1 / (1 + self.metric(X, self.mean_))



from __future__ import annotations

from typing import Callable

import scipy.integrate

from ...representation import FData
from ...typing._numpy import NDArrayFloat
from ._operators import Operator


class IntegralTransform(
    Operator[FData, Callable[[NDArrayFloat], NDArrayFloat]],
):
    """Integral operator.

    Parameters:
        kernel_function:  Kernel function corresponding to the operator.

    """

    def __init__(
        self,
        kernel_function: Callable[[NDArrayFloat, NDArrayFloat], NDArrayFloat],
    ) -> None:
        self.kernel_function = kernel_function

    def __call__(  # noqa: D102
        self,
        f: FData,
    ) -> Callable[[NDArrayFloat], NDArrayFloat]:

        def evaluate_covariance(  # noqa: WPS430
            points: NDArrayFloat,
        ) -> NDArrayFloat:

            def integral_body(  # noqa: WPS430
                integration_var: NDArrayFloat,
            ) -> NDArrayFloat:
                return (
                    f(integration_var)
                    * self.kernel_function(integration_var, points)
                )

            domain_range = f.domain_range[0]

            return scipy.integrate.quad_vec(  # type: ignore[no-any-return]
                integral_body,
                domain_range[0],
                domain_range[1],
            )[0]

        return evaluate_covariance



"""Miscellaneous functions and objects."""
from typing import TYPE_CHECKING

import lazy_loader as lazy

__getattr__, __dir__, __all__ = lazy.attach(
    __name__,
    submodules=[
        "covariances",
        "hat_matrix",
        "kernels",
        "lstsq",
        "metrics",
        "operators",
        "regularization",
        "scoring",
        "validation",
    ],
    submod_attrs={
        '_math': [
            "cosine_similarity",
            "cosine_similarity_matrix",
            "cumsum",
            "exp",
            "inner_product",
            "inner_product_matrix",
            "log",
            "log2",
            "log10",
            "sqrt",
        ],
    },
)

if TYPE_CHECKING:

    from ._math import (
        cosine_similarity as cosine_similarity,
        cosine_similarity_matrix as cosine_similarity_matrix,
        cumsum as cumsum,
        exp as exp,
        inner_product as inner_product,
        inner_product_matrix as inner_product_matrix,
        log as log,
        log2 as log2,
        log10 as log10,
        sqrt as sqrt,
    )



"""Test to check the Fda Feature Union module."""

import unittest

import pandas as pd
from pandas.testing import assert_frame_equal

from skfda.datasets import fetch_growth
from skfda.misc.hat_matrix import NadarayaWatsonHatMatrix
from skfda.misc.operators import SRSF
from skfda.preprocessing.feature_construction import (
    EvaluationTransformer,
    FDAFeatureUnion,
)
from skfda.preprocessing.smoothing import KernelSmoother


class TestFDAFeatureUnion(unittest.TestCase):
    """Check the Fda Feature Union module."""

    def setUp(self) -> None:
        """Fetch the Berkeley Growth Study dataset."""
        self.X = fetch_growth(return_X_y=True)[0]

    def test_incompatible_fdatagrid_output(self) -> None:
        """Check that array_output fails for FData features."""
        u = FDAFeatureUnion(
            [("eval", EvaluationTransformer(None)), ("srsf", SRSF())],
            array_output=True,
        )

        with self.assertRaises(TypeError):
            u.fit_transform(self.X)

    def test_correct_transformation_concat(self) -> None:
        """Check that the transformation is done correctly."""
        u = FDAFeatureUnion(
            [
                ("srsf1", SRSF()),
                (
                    "smooth",
                    KernelSmoother(
                        kernel_estimator=NadarayaWatsonHatMatrix(),
                    ),
                ),
            ],
        )
        created_frame = u.fit_transform(self.X)

        t1 = SRSF().fit_transform(self.X)
        t2 = KernelSmoother(
            kernel_estimator=NadarayaWatsonHatMatrix(),
        ).fit_transform(self.X)

        true_frame = pd.concat(
            [
                pd.DataFrame({0: t1}),
                pd.DataFrame({0: t2}),
            ],
            axis=1,
        )
        assert_frame_equal(true_frame, created_frame)


if __name__ == '__main__':
    unittest.main()



"""Scoring methods for FData."""
from __future__ import annotations

import math
import warnings
from functools import singledispatch
from typing import Callable, Optional, TypeVar, Union, overload

import numpy as np
import sklearn.metrics
from typing_extensions import Literal, Protocol

from .._utils import nquad_vec
from ..representation import FData, FDataBasis, FDataGrid
from ..representation._functional_data import EvalPointsType
from ..typing._numpy import NDArrayFloat

DataType = TypeVar('DataType')

MultiOutputType = Literal['uniform_average', 'raw_values']


class _InfiniteScoreError(Exception):
    """Exception for skipping integral on infinite value."""


class ScoreFunction(Protocol):
    """Type definition for score functions."""

    @overload
    def __call__(
        self,
        y_true: DataType,
        y_pred: DataType,
        *,
        sample_weight: NDArrayFloat | None = None,
        multioutput: Literal['uniform_average'] = 'uniform_average',
    ) -> float:
        pass  # noqa: WPS428

    @overload
    def __call__(   # noqa: D102
        self,
        y_true: DataType,
        y_pred: DataType,
        *,
        sample_weight: NDArrayFloat | None = None,
        multioutput: Literal['raw_values'],
    ) -> DataType:
        pass  # noqa: WPS428

    def __call__(   # noqa: D102
        self,
        y_true: DataType,
        y_pred: DataType,
        *,
        sample_weight: NDArrayFloat | None = None,
        multioutput: MultiOutputType = 'uniform_average',
    ) -> float | DataType:
        pass  # noqa: WPS428


def _domain_measure(fd: FData) -> float:
    measure = 1.0
    for interval in fd.domain_range:
        measure = measure * (interval[1] - interval[0])
    return measure


def _var(
    x: FDataGrid,
    weights: NDArrayFloat | None = None,
) -> FDataGrid:
    from ..exploratory.stats import mean, var

    if weights is None:
        return var(x, correction=0)

    return mean(  # type: ignore[no-any-return]
        np.power(x - mean(x, weights=weights), 2),
        weights=weights,
    )


def _multioutput_score_basis(
    y_true: FDataBasis,
    multioutput: MultiOutputType,
    integrand: Callable[[NDArrayFloat], NDArrayFloat],
) -> float:

    if multioutput != "uniform_average":
        raise ValueError(
            f"Only \"uniform_average\" is supported for \"multioutput\" when "
            f"the input is a FDatabasis: received {multioutput} instead",
        )

    try:
        integral = nquad_vec(
            integrand,
            y_true.domain_range,
        )
    except _InfiniteScoreError:
        return -math.inf

    # If the dimension of the codomain is > 1,
    # the mean of the scores is taken
    return float(np.mean(integral) / _domain_measure(y_true))


def _multioutput_score_grid(
    score: FDataGrid,
    multioutput: MultiOutputType,
    squared: bool = True,
) -> float | FDataGrid:

    if not squared:
        score = np.sqrt(score)

    if multioutput == 'raw_values':
        return score

    # Score only contains 1 function
    # If the dimension of the codomain is > 1,
    # the mean of the scores is taken
    return float(np.mean(score.integrate()[0]) / _domain_measure(score))


@overload
def explained_variance_score(
    y_true: DataType,
    y_pred: DataType,
    *,
    sample_weight: NDArrayFloat | None = None,
    multioutput: Literal['uniform_average'] = 'uniform_average',
) -> float:
    pass  # noqa: WPS428


@overload
def explained_variance_score(
    y_true: DataType,
    y_pred: DataType,
    *,
    sample_weight: NDArrayFloat | None = None,
    multioutput: Literal['raw_values'],
) -> DataType:
    pass  # noqa: WPS428


@singledispatch
def explained_variance_score(
    y_true: DataType,
    y_pred: DataType,
    *,
    sample_weight: NDArrayFloat | None = None,
    multioutput: MultiOutputType = 'uniform_average',
) -> float | DataType:
    r"""Explained variance score for :class:`~skfda.representation.FData`.

    With :math:`y\_true = (X_1, X_2, ..., X_n)` being the real values,
    :math:`t\_pred = (\hat{X}_1, \hat{X}_2, ..., \hat{X}_n)` being the
    estimated and :math:`sample\_weight = (w_1, w_2, ..., w_n)`, the score is
    calculated as

    .. math::
        EV(y\_true, y\_pred)(t) = 1 -
        \frac{Var(y\_true(t) - y\_pred(t), sample\_weight)}
        {Var(y\_true(t), sample\_weight)}

    where :math:`Var` is a weighted variance.

    Weighted variance is defined as below

    .. math::
        Var(y\_true, sample\_weight)(t) = \sum_{i=1}^n w_i
        (X_i(t) - Mean(fd(t), sample\_weight))^2.

    Here, :math:`Mean` is a weighted mean.

    For :math:`y\_true` and :math:`y\_pred` of type
    :class:`~skfda.representation.FDataGrid`,  :math:`EV` is
    also a :class:`~skfda.representation.FDataGrid` object with
    the same grid points.

    If multioutput = 'raw_values', the function :math:`EV` is returned.
    Otherwise, if multioutput = 'uniform_average', the mean of :math:`EV` is
    calculated:

    .. math::
        mean(EV) = \frac{1}{V}\int_{D} EV(t) dt

    where :math:`D` is the function domain and :math:`V` the volume of that
    domain.

    For :class:`~skfda.representation.FDataBasis` only
    'uniform_average' is available.

    If :math:`y\_true` and :math:`y\_pred` are numpy arrays, sklearn function
    is called.

    The best possible score is 1.0, lower values are worse.

    Args:
        y_true: Correct target values.
        y_pred: Estimated values.
        sample_weight: Sample weights. By default, uniform weights
            are taken.
        multioutput: Defines format of the return.

    Returns:
        Explained variance score.

        If multioutput = 'uniform_average' or
        :math:`y\_pred` and :math:`y\_true` are
        :class:`~skfda.representation.FDataBasis` objects, float is returned.

        If both :math:`y\_pred` and :math:`y\_true` are
        :class:`~skfda.representation.FDataGrid`
        objects and multioutput = 'raw_values',
        :class:`~skfda.representation.FDataGrid` is returned.

        If both :math:`y\_pred` and :math:`y\_true` are ndarray and
        multioutput = 'raw_values', ndarray.

    """
    return (  # type: ignore [no-any-return]
        sklearn.metrics.explained_variance_score(
            y_true,
            y_pred,
            sample_weight=sample_weight,
            multioutput=multioutput,
        )
    )


@explained_variance_score.register  # type: ignore[attr-defined, misc]
def _explained_variance_score_fdatagrid(
    y_true: FDataGrid,
    y_pred: FDataGrid,
    *,
    sample_weight: Optional[NDArrayFloat] = None,
    multioutput: MultiOutputType = 'uniform_average',
) -> Union[float, FDataGrid]:

    num = _var(y_true - y_pred, weights=sample_weight)
    den = _var(y_true, weights=sample_weight)

    # Divisions by zero allowed
    with np.errstate(divide='ignore', invalid='ignore'):
        score = 1 - num / den

    # 0 / 0 divisions should be 0 in this context, and the score, 1
    score.data_matrix[np.isnan(score.data_matrix)] = 1

    return _multioutput_score_grid(score, multioutput)


@explained_variance_score.register  # type: ignore[attr-defined, misc]
def _explained_variance_score_fdatabasis(
    y_true: FDataBasis,
    y_pred: FDataBasis,
    *,
    sample_weight: Optional[NDArrayFloat] = None,
    multioutput: MultiOutputType = 'uniform_average',
) -> float:

    def _ev_func(x: EvalPointsType) -> NDArrayFloat:  # noqa: WPS430
        num = np.average(
            (
                (y_true(x) - y_pred(x))
                - np.average(
                    y_true(x) - y_pred(x),
                    weights=sample_weight,
                    axis=0,
                )
            ) ** 2,
            weights=sample_weight,
            axis=0,
        )

        den = np.average(
            (
                y_true(x)
                - np.average(y_true(x), weights=sample_weight, axis=0)
            ) ** 2,
            weights=sample_weight,
            axis=0,
        )

        # Divisions by zero allowed
        with np.errstate(divide='ignore', invalid='ignore'):
            score = 1 - num / den

        # 0/0 case, the score is 1.
        score[np.isnan(score)] = 1

        # r/0 case, r!= 0. Return -inf outside this function
        if np.any(np.isinf(score)):
            raise _InfiniteScoreError

        # Score only contains 1 input point
        assert score.shape[0] == 1
        return score[0]  # type: ignore [no-any-return]

    return _multioutput_score_basis(y_true, multioutput, _ev_func)


@overload
def mean_absolute_error(
    y_true: DataType,
    y_pred: DataType,
    *,
    sample_weight: NDArrayFloat | None = None,
    multioutput: Literal['uniform_average'] = 'uniform_average',
) -> float:
    pass  # noqa: WPS428


@overload
def mean_absolute_error(
    y_true: DataType,
    y_pred: DataType,
    *,
    sample_weight: NDArrayFloat | None = None,
    multioutput: Literal['raw_values'],
) -> DataType:
    pass  # noqa: WPS428


@singledispatch
def mean_absolute_error(
    y_true: DataType,
    y_pred: DataType,
    *,
    sample_weight: NDArrayFloat | None = None,
    multioutput: MultiOutputType = 'uniform_average',
) -> float | DataType:
    r"""Mean Absolute Error for :class:`~skfda.representation.FData`.

    With :math:`y\_true = (X_1, X_2, ..., X_n)` being the real values,
    :math:`t\_pred = (\hat{X}_1, \hat{X}_2, ..., \hat{X}_n)` being the
    estimated and :math:`sample\_weight = (w_1, w_2, ..., w_n)`, the error is
    calculated as

    .. math::
        MAE(y\_true, y\_pred)(t) = \frac{1}{\sum w_i}
        \sum_{i=1}^n w_i|X_i(t) - \hat{X}_i(t)|

    For :math:`y\_true` and :math:`y\_pred` of type
    :class:`~skfda.representation.FDataGrid`,  :math:`MAE` is
    also a :class:`~skfda.representation.FDataGrid` object with
    the same grid points.

    If multioutput = 'raw_values', the function :math:`MAE` is returned.
    Otherwise, if multioutput = 'uniform_average', the mean of :math:`MAE` is
    calculated:

    .. math::
        mean(MAE) = \frac{1}{V}\int_{D} MAE(t) dt

    where :math:`D` is the function domain and :math:`V` the volume of that
    domain.

    For :class:`~skfda.representation.FDataBasis` only
    'uniform_average' is available.

    If :math:`y\_true` and :math:`y\_pred` are numpy arrays, sklearn function
    is called.

    Args:
        y_true: Correct target values.
        y_pred: Estimated values.
        sample_weight: Sample weights. By default, uniform weights
            are taken.
        multioutput: Defines format of the return.

    Returns:
        Mean absolute error.

        If multioutput = 'uniform_average' or
        :math:`y\_pred` and :math:`y\_true` are
        :class:`~skfda.representation.FDataBasis` objects, float is returned.

        If both :math:`y\_pred` and :math:`y\_true` are
        :class:`~skfda.representation.FDataGrid`
        objects and multioutput = 'raw_values',
        :class:`~skfda.representation.FDataGrid` is returned.

        If both :math:`y\_pred` and :math:`y\_true` are ndarray and
        multioutput = 'raw_values', ndarray.

    """
    return sklearn.metrics.mean_absolute_error(  # type: ignore [no-any-return]
        y_true,
        y_pred,
        sample_weight=sample_weight,
        multioutput=multioutput,
    )


@mean_absolute_error.register  # type: ignore[attr-defined, misc]
def _mean_absolute_error_fdatagrid(
    y_true: FDataGrid,
    y_pred: FDataGrid,
    *,
    sample_weight: Optional[NDArrayFloat] = None,
    multioutput: MultiOutputType = 'uniform_average',
) -> Union[float, FDataGrid]:
    from ..exploratory.stats import mean

    error = mean(np.abs(y_true - y_pred), weights=sample_weight)
    return _multioutput_score_grid(error, multioutput)


@mean_absolute_error.register  # type: ignore[attr-defined, misc]
def _mean_absolute_error_fdatabasis(
    y_true: FDataBasis,
    y_pred: FDataBasis,
    *,
    sample_weight: Optional[NDArrayFloat] = None,
    multioutput: MultiOutputType = 'uniform_average',
) -> float:

    def _mae_func(x: EvalPointsType) -> NDArrayFloat:  # noqa: WPS430
        error = np.average(
            np.abs(y_true(x) - y_pred(x)),
            weights=sample_weight,
            axis=0,
        )

        # Error only contains 1 input point
        assert error.shape[0] == 1
        return error[0]  # type: ignore [no-any-return]

    return _multioutput_score_basis(y_true, multioutput, _mae_func)


@overload
def mean_absolute_percentage_error(
    y_true: DataType,
    y_pred: DataType,
    *,
    sample_weight: NDArrayFloat | None = None,
    multioutput: Literal['uniform_average'] = 'uniform_average',
) -> float:
    pass  # noqa: WPS428


@overload
def mean_absolute_percentage_error(
    y_true: DataType,
    y_pred: DataType,
    *,
    sample_weight: NDArrayFloat | None = None,
    multioutput: Literal['raw_values'],
) -> DataType:
    pass  # noqa: WPS428


@singledispatch
def mean_absolute_percentage_error(
    y_true: DataType,
    y_pred: DataType,
    *,
    sample_weight: NDArrayFloat | None = None,
    multioutput: MultiOutputType = 'uniform_average',
) -> float | DataType:
    r"""Mean Absolute Percentage Error for :class:`~skfda.representation.FData`.

    With :math:`y\_true = (X_1, X_2, ..., X_n)` being the real values,
    :math:`t\_pred = (\hat{X}_1, \hat{X}_2, ..., \hat{X}_n)` being the
    estimated and :math:`sample\_weight = (w_1, w_2, ..., w_n)`, the error is
    calculated as

    .. math::
        MAPE(y\_true, y\_pred)(t) = \frac{1}{\sum w_i}
        \sum_{i=1}^n w_i\frac{|X_i(t) - \hat{X}_i(t)|}{|X_i(t)|}

    For :math:`y\_true` and :math:`y\_pred` of type
    :class:`~skfda.representation.FDataGrid`,  :math:`MAPE` is
    also a :class:`~skfda.representation.FDataGrid` object with
    the same grid points.

    If multioutput = 'raw_values', the function :math:`MAPE` is returned.
    Otherwise, if multioutput = 'uniform_average', the mean of :math:`MAPE` is
    calculated:

    .. math::
        mean(MAPE) = \frac{1}{V}\int_{D} MAPE(t) dt

    where :math:`D` is the function domain and :math:`V` the volume of that
    domain.

    For :class:`~skfda.representation.FDataBasis` only
    'uniform_average' is available.

    If :math:`y\_true` and :math:`y\_pred` are numpy arrays, sklearn function
    is called.

    This function should not be used if for some :math:`t` and some :math:`i`,
    :math:`X_i(t) = 0`.

    Args:
        y_true: Correct target values.
        y_pred: Estimated values.
        sample_weight: Sample weights. By default, uniform weights
            are taken.
        multioutput: Defines format of the return.

    Returns:
        Mean absolute percentage error.

        If multioutput = 'uniform_average' or
        :math:`y\_pred` and :math:`y\_true` are
        :class:`~skfda.representation.FDataBasis` objects, float is returned.

        If both :math:`y\_pred` and :math:`y\_true` are
        :class:`~skfda.representation.FDataGrid`
        objects and multioutput = 'raw_values',
        :class:`~skfda.representation.FDataGrid` is returned.

        If both :math:`y\_pred` and :math:`y\_true` are ndarray and
        multioutput = 'raw_values', ndarray.

    """
    return (  # type: ignore [no-any-return]
        sklearn.metrics.mean_absolute_percentage_error(
            y_true,
            y_pred,
            sample_weight=sample_weight,
            multioutput=multioutput,
        )
    )


@mean_absolute_percentage_error.register  # type: ignore[attr-defined, misc]
def _mean_absolute_percentage_error_fdatagrid(
    y_true: FDataGrid,
    y_pred: FDataGrid,
    *,
    sample_weight: Optional[NDArrayFloat] = None,
    multioutput: MultiOutputType = 'uniform_average',
) -> Union[float, FDataGrid]:
    from ..exploratory.stats import mean

    epsilon = np.finfo(np.float64).eps

    if np.any(np.abs(y_true.data_matrix) < epsilon):
        warnings.warn('Zero denominator', RuntimeWarning)

    mape = np.abs(y_pred - y_true) / np.maximum(np.abs(y_true), epsilon)

    error = mean(mape, weights=sample_weight)
    return _multioutput_score_grid(error, multioutput)


@mean_absolute_percentage_error.register  # type: ignore[attr-defined, misc]
def _mean_absolute_percentage_error_fdatabasis(
    y_true: FDataBasis,
    y_pred: FDataBasis,
    *,
    sample_weight: Optional[NDArrayFloat] = None,
    multioutput: MultiOutputType = 'uniform_average',
) -> float:

    def _mape_func(x: EvalPointsType) -> NDArrayFloat:  # noqa: WPS430

        epsilon = np.finfo(np.float64).eps
        if np.any(np.abs(y_true(x)) < epsilon):
            warnings.warn('Zero denominator', RuntimeWarning)

        error = np.average(
            (
                np.abs(y_true(x) - y_pred(x))
                / np.maximum(np.abs(y_true(x)), epsilon)
            ),
            weights=sample_weight,
            axis=0,
        )

        # Error only contains 1 input point
        assert error.shape[0] == 1
        return error[0]  # type: ignore [no-any-return]

    return _multioutput_score_basis(y_true, multioutput, _mape_func)


@overload
def mean_squared_error(
    y_true: DataType,
    y_pred: DataType,
    *,
    sample_weight: NDArrayFloat | None = None,
    multioutput: Literal['uniform_average'] = 'uniform_average',
    squared: bool = True,
) -> float:
    pass  # noqa: WPS428


@overload
def mean_squared_error(
    y_true: DataType,
    y_pred: DataType,
    *,
    sample_weight: NDArrayFloat | None = None,
    multioutput: Literal['raw_values'],
    squared: bool = True,
) -> DataType:
    pass  # noqa: WPS428


@singledispatch
def mean_squared_error(
    y_true: DataType,
    y_pred: DataType,
    *,
    sample_weight: NDArrayFloat | None = None,
    multioutput: MultiOutputType = 'uniform_average',
    squared: bool = True,
) -> float | DataType:
    r"""Mean Squared Error for :class:`~skfda.representation.FData`.

    With :math:`y\_true = (X_1, X_2, ..., X_n)` being the real values,
    :math:`t\_pred = (\hat{X}_1, \hat{X}_2, ..., \hat{X}_n)` being the
    estimated and :math:`sample\_weight = (w_1, w_2, ..., w_n)`, the error is
    calculated as

    .. math::
        MSE(y\_true, y\_pred)(t) = \frac{1}{\sum w_i}
        \sum_{i=1}^n w_i(X_i(t) - \hat{X}_i(t))^2

    For :math:`y\_true` and :math:`y\_pred` of type
    :class:`~skfda.representation.FDataGrid`,  :math:`MSE` is
    also a :class:`~skfda.representation.FDataGrid` object with
    the same grid points.

    If multioutput = 'raw_values', the function :math:`MSE` is returned.
    Otherwise, if multioutput = 'uniform_average', the mean of :math:`MSE` is
    calculated:

    .. math::
        mean(MSE) = \frac{1}{V}\int_{D} MSE(t) dt

    where :math:`D` is the function domain and :math:`V` the volume of that
    domain.

    For :class:`~skfda.representation.FDataBasis` only
    'uniform_average' is available.

    If :math:`y\_true` and :math:`y\_pred` are numpy arrays, sklearn function
    is called.

    Args:
        y_true: Correct target values.
        y_pred: Estimated values.
        sample_weight: Sample weights. By default, uniform weights
            are taken.
        multioutput: Defines format of the return.
        squared: If True returns MSE value, if False returns RMSE value.

    Returns:
        Mean squared error.

        If multioutput = 'uniform_average' or
        :math:`y\_pred` and :math:`y\_true` are
        :class:`~skfda.representation.FDataBasis` objects, float is returned.

        If both :math:`y\_pred` and :math:`y\_true` are
        :class:`~skfda.representation.FDataGrid`
        objects and multioutput = 'raw_values',
        :class:`~skfda.representation.FDataGrid` is returned.

        If both :math:`y\_pred` and :math:`y\_true` are ndarray and
        multioutput = 'raw_values', ndarray.

    """
    return sklearn.metrics.mean_squared_error(  # type: ignore [no-any-return]
        y_true,
        y_pred,
        sample_weight=sample_weight,
        multioutput=multioutput,
        squared=squared,
    )


@mean_squared_error.register  # type: ignore[attr-defined, misc]
def _mean_squared_error_fdatagrid(
    y_true: FDataGrid,
    y_pred: FDataGrid,
    *,
    sample_weight: Optional[NDArrayFloat] = None,
    multioutput: MultiOutputType = 'uniform_average',
    squared: bool = True,
) -> Union[float, FDataGrid]:
    from ..exploratory.stats import mean

    error: FDataGrid = mean(
        np.power(y_true - y_pred, 2),
        weights=sample_weight,
    )

    return _multioutput_score_grid(error, multioutput, squared=squared)


@mean_squared_error.register  # type: ignore[attr-defined, misc]
def _mean_squared_error_fdatabasis(
    y_true: FDataBasis,
    y_pred: FDataBasis,
    *,
    sample_weight: Optional[NDArrayFloat] = None,
    squared: bool = True,
    multioutput: MultiOutputType = 'uniform_average',
) -> float:

    def _mse_func(x: EvalPointsType) -> NDArrayFloat:  # noqa: WPS430

        error: NDArrayFloat = np.average(
            (y_true(x) - y_pred(x)) ** 2,
            weights=sample_weight,
            axis=0,
        )

        if not squared:
            return np.sqrt(error)

        # Error only contains 1 input point
        assert error.shape[0] == 1
        return error[0]  # type: ignore [no-any-return]

    return _multioutput_score_basis(y_true, multioutput, _mse_func)


@overload
def mean_squared_log_error(
    y_true: DataType,
    y_pred: DataType,
    *,
    sample_weight: NDArrayFloat | None = None,
    multioutput: Literal['uniform_average'] = 'uniform_average',
    squared: bool = True,
) -> float:
    pass  # noqa: WPS428


@overload
def mean_squared_log_error(
    y_true: DataType,
    y_pred: DataType,
    *,
    sample_weight: NDArrayFloat | None = None,
    multioutput: Literal['raw_values'],
    squared: bool = True,
) -> DataType:
    pass  # noqa: WPS428


@singledispatch
def mean_squared_log_error(
    y_true: DataType,
    y_pred: DataType,
    *,
    sample_weight: NDArrayFloat | None = None,
    multioutput: MultiOutputType = 'uniform_average',
    squared: bool = True,
) -> float | DataType:
    r"""Mean Squared Log Error for :class:`~skfda.representation.FData`.

    With :math:`y\_true = (X_1, X_2, ..., X_n)` being the real values,
    :math:`t\_pred = (\hat{X}_1, \hat{X}_2, ..., \hat{X}_n)` being the
    estimated and :math:`sample\_weight = (w_1, w_2, ..., w_n)`, the error is
    calculated as

    .. math::
        MSLE(y\_true, y\_pred)(t) = \frac{1}{\sum w_i}
        \sum_{i=1}^n w_i(\log(1 + X_i(t)) - \log(1 + \hat{X}_i(t)))^2

    where :math:`\log` is the natural logarithm.

    For :math:`y\_true` and :math:`y\_pred` of type
    :class:`~skfda.representation.FDataGrid`,  :math:`MSLE` is
    also a :class:`~skfda.representation.FDataGrid` object with
    the same grid points.

    If multioutput = 'raw_values', the function :math:`MSLE` is returned.
    Otherwise, if multioutput = 'uniform_average', the mean of :math:`MSLE` is
    calculated:

    .. math::
        mean(MSLE) = \frac{1}{V}\int_{D} MSLE(t) dt

    where :math:`D` is the function domain and :math:`V` the volume of that
    domain.

    For :class:`~skfda.representation.FDataBasis` only
    'uniform_average' is available.

    If :math:`y\_true` and :math:`y\_pred` are numpy arrays, sklearn function
    is called.

    This function should not be used if for some :math:`t` and some :math:`i`,
    :math:`X_i(t) < 0`.

    Args:
        y_true: Correct target values.
        y_pred: Estimated values.
        sample_weight: Sample weights. By default, uniform weights
            are taken.
        multioutput: Defines format of the return.
        squared: default True. If False, square root is taken.

    Returns:
        Mean squared log error.

        If multioutput = 'uniform_average' or
        :math:`y\_pred` and :math:`y\_true` are
        :class:`~skfda.representation.FDataBasis` objects, float is returned.

        If both :math:`y\_pred` and :math:`y\_true` are
        :class:`~skfda.representation.FDataGrid`
        objects and multioutput = 'raw_values',
        :class:`~skfda.representation.FDataGrid` is returned.

        If both :math:`y\_pred` and :math:`y\_true` are ndarray and
        multioutput = 'raw_values', ndarray.

    """
    return (  # type: ignore [no-any-return]
        sklearn.metrics.mean_squared_log_error(
            y_true,
            y_pred,
            sample_weight=sample_weight,
            multioutput=multioutput,
            squared=squared,
        )
    )


@mean_squared_log_error.register  # type: ignore[attr-defined, misc]
def _mean_squared_log_error_fdatagrid(
    y_true: FDataGrid,
    y_pred: FDataGrid,
    *,
    sample_weight: Optional[NDArrayFloat] = None,
    multioutput: MultiOutputType = 'uniform_average',
    squared: bool = True,
) -> Union[float, FDataGrid]:

    if np.any(y_true.data_matrix < 0) or np.any(y_pred.data_matrix < 0):
        raise ValueError(
            "Mean Squared Logarithmic Error cannot be used when "
            "targets functions have negative values.",
        )

    return mean_squared_error(
        np.log1p(y_true),
        np.log1p(y_pred),
        sample_weight=sample_weight,
        multioutput=multioutput,
        squared=squared,
    )


@mean_squared_log_error.register  # type: ignore[attr-defined, misc]
def _mean_squared_log_error_fdatabasis(
    y_true: FDataBasis,
    y_pred: FDataBasis,
    *,
    sample_weight: Optional[NDArrayFloat] = None,
    squared: bool = True,
    multioutput: MultiOutputType = 'uniform_average',
) -> float:

    def _msle_func(x: EvalPointsType) -> NDArrayFloat:  # noqa: WPS430

        y_true_eval = y_true(x)
        y_pred_eval = y_pred(x)

        if np.any(y_true_eval < 0) or np.any(y_pred_eval < 0):
            raise ValueError(
                "Mean Squared Logarithmic Error cannot be used when "
                "targets functions have negative values.",
            )

        error: NDArrayFloat = np.average(
            (np.log1p(y_true_eval) - np.log1p(y_pred_eval)) ** 2,
            weights=sample_weight,
            axis=0,
        )

        if not squared:
            return np.sqrt(error)

        # Error only contains 1 input point
        assert error.shape[0] == 1
        return error[0]  # type: ignore [no-any-return]

    return _multioutput_score_basis(y_true, multioutput, _msle_func)


@overload
def r2_score(
    y_true: DataType,
    y_pred: DataType,
    *,
    sample_weight: NDArrayFloat | None = None,
    multioutput: Literal['uniform_average'] = 'uniform_average',
) -> float:
    pass  # noqa: WPS428


@overload
def r2_score(
    y_true: DataType,
    y_pred: DataType,
    *,
    sample_weight: NDArrayFloat | None = None,
    multioutput: Literal['raw_values'],
) -> DataType:
    pass  # noqa: WPS428


@singledispatch
def r2_score(
    y_true: DataType,
    y_pred: DataType,
    *,
    sample_weight: NDArrayFloat | None = None,
    multioutput: MultiOutputType = 'uniform_average',
) -> float | DataType:
    r"""R^2 score for :class:`~skfda.representation.FData`.

    With :math:`y\_true = (X_1, X_2, ..., X_n)` being the real values,
    :math:`t\_pred = (\hat{X}_1, \hat{X}_2, ..., \hat{X}_n)` being the
    estimated and :math:`sample\_weight = (w_1, w_2, ..., w_n)`, the score is
    calculated as

    .. math::
        R^2(y\_true, y\_pred)(t) = 1 -
        \frac{\sum_{i=1}^n w_i (X_i(t) - \hat{X}_i(t))^2}
        {\sum_{i=1}^n w_i (X_i(t) - Mean(y\_true, sample\_weight)(t))^2}

    where :math:`Mean` is a weighted mean.

    For :math:`y\_true` and :math:`y\_pred` of type
    :class:`~skfda.representation.FDataGrid`,  :math:`R^2` is
    also a :class:`~skfda.representation.FDataGrid` object with
    the same grid points.

    If multioutput = 'raw_values', the function :math:`R^2` is returned.
    Otherwise, if multioutput = 'uniform_average', the mean of :math:`R^2` is
    calculated:

    .. math::
        mean(R^2) = \frac{1}{V}\int_{D} R^2(t) dt

    where :math:`D` is the function domain and :math:`V` the volume of that
    domain.

    For :class:`~skfda.representation.FDataBasis` only
    'uniform_average' is available.

    If :math:`y\_true` and :math:`y\_pred` are numpy arrays, sklearn function
    is called.

    Args:
        y_true: Correct target values.
        y_pred: Estimated values.
        sample_weight: Sample weights. By default, uniform weights
            are taken.
        multioutput: Defines format of the return.

    Returns:
        R2 score

        If multioutput = 'uniform_average' or
        :math:`y\_pred` and :math:`y\_true` are
        :class:`~skfda.representation.FDataBasis` objects, float is returned.

        If both :math:`y\_pred` and :math:`y\_true` are
        :class:`~skfda.representation.FDataGrid`
        objects and multioutput = 'raw_values',
        :class:`~skfda.representation.FDataGrid` is returned.

        If both :math:`y\_pred` and :math:`y\_true` are ndarray and
        multioutput = 'raw_values', ndarray.

    """
    return sklearn.metrics.r2_score(  # type: ignore [no-any-return]
        y_true,
        y_pred,
        sample_weight=sample_weight,
        multioutput=multioutput,
    )


@r2_score.register  # type: ignore[attr-defined, misc]
def _r2_score_fdatagrid(
    y_true: FDataGrid,
    y_pred: FDataGrid,
    *,
    sample_weight: Optional[NDArrayFloat] = None,
    multioutput: MultiOutputType = 'uniform_average',
) -> Union[float, FDataGrid]:
    from ..exploratory.stats import mean

    if y_pred.n_samples < 2:
        raise ValueError(
            'R^2 score is not well-defined with less than two samples.',
        )

    ss_res = mean(
        np.power(y_true - y_pred, 2),
        weights=sample_weight,
    )

    ss_tot = _var(y_true, weights=sample_weight)

    # Divisions by zero allowed
    with np.errstate(divide='ignore', invalid='ignore'):
        score: FDataGrid = 1 - ss_res / ss_tot

    # 0 / 0 divisions should be 0 in this context and the score, 1
    score.data_matrix[np.isnan(score.data_matrix)] = 1

    return _multioutput_score_grid(score, multioutput)


@r2_score.register  # type: ignore[attr-defined, misc]
def _r2_score_fdatabasis(
    y_true: FDataBasis,
    y_pred: FDataBasis,
    *,
    sample_weight: Optional[NDArrayFloat] = None,
    multioutput: MultiOutputType = 'uniform_average',
) -> float:

    if y_pred.n_samples < 2:
        raise ValueError(
            'R^2 score is not well-defined with less than two samples.',
        )

    def _r2_func(x: NDArrayFloat) -> NDArrayFloat:  # noqa: WPS430
        ss_res = np.average(
            (y_true(x) - y_pred(x)) ** 2,
            weights=sample_weight,
            axis=0,
        )

        ss_tot = np.average(
            (
                y_true(x)
                - np.average(y_true(x), weights=sample_weight, axis=0)
            ) ** 2,
            weights=sample_weight,
            axis=0,
        )

        # Divisions by zero allowed
        with np.errstate(divide='ignore', invalid='ignore'):
            score = 1 - ss_res / ss_tot

        # 0/0 case, the score is 1.
        score[np.isnan(score)] = 1

        # r/0 case, r!= 0. Return -inf outside this function
        if np.any(np.isinf(score)):
            raise _InfiniteScoreError

        # Score only had 1 input point
        assert score.shape[0] == 1
        return score[0]  # type: ignore [no-any-return]

    return _multioutput_score_basis(y_true, multioutput, _r2_func)



from __future__ import annotations

from typing import Callable, Tuple

import numpy as np
from sklearn.linear_model import LogisticRegression as mvLogisticRegression
from sklearn.utils.validation import check_is_fitted
from typing_extensions import Literal

from ..._utils import _classifier_get_classes
from ..._utils._sklearn_adapter import BaseEstimator, ClassifierMixin
from ...representation import FDataGrid
from ...typing._numpy import NDArrayAny, NDArrayInt

Solver = Literal["newton-cg", "lbfgs", "liblinear", "sag", "saga"]


class LogisticRegression(
    BaseEstimator,
    ClassifierMixin[FDataGrid, NDArrayAny],
):
    r"""Logistic Regression classifier for functional data.

    This class implements the sequential “greedy” algorithm
    for functional logistic regression proposed in
    :footcite:ts:`berrendero++_2022_functional`.

    .. warning::
        For now, only binary classification for functional
        data with one dimensional domains is supported.

    Args:
        n_features_to_select:
            Number of points (and coefficients) to be selected by
            the algorithm.
        penalty:
            Penalty to use in the multivariate logistic regresion
            optimization problem. For more info check the parameter
            "penalty" in
            :external:class:`sklearn.linear_model.LogisticRegression`.
        C:
            Inverse of the regularization parameter in the multivariate
            logistic regresion optimization problem. For more info
            check the parameter "C" in
            :external:class:`sklearn.linear_model.LogisticRegression`.
        solver:
            Algorithm to use in the multivariate logistic regresion
            optimization problem. For more info check the parameter
            "solver" in
            :external:class:`sklearn.linear_model.LogisticRegression`.
        max_iter:
            Maximum number of iterations taken for the solver to converge.

    Attributes:
        classes\_: A list containing the name of the classes
        points\_: A list containing the selected points.
        coef\_: A list containing the coefficient for each selected point.
        intercept\_: Independent term.

    Examples:
        >>> import skfda
        >>> from skfda.datasets import make_gaussian_process
        >>> from skfda.ml.classification import LogisticRegression
        >>>
        >>> n_samples = 10000
        >>> n_features = 200
        >>>
        >>> def mean_1(t):
        ...     return (np.abs(t - 0.25)
        ...             - 2 * np.abs(t - 0.5)
        ...             + np.abs(t - 0.75))
        >>>
        >>> X_0 = make_gaussian_process(
        ...     n_samples=n_samples // 2,
        ...     n_features=n_features,
        ...     random_state=0,
        ... )
        >>> X_1 = make_gaussian_process(
        ...     n_samples=n_samples // 2,
        ...     n_features=n_features,
        ...     mean=mean_1,
        ...     random_state=1,
        ... )
        >>> X = skfda.concatenate((X_0, X_1))
        >>>
        >>> y = np.zeros(n_samples)
        >>> y [n_samples // 2:] = 1
        >>> lr = LogisticRegression(max_features=3)
        >>> _ = lr.fit(X[::2], y[::2])
        >>> np.allclose(sorted(lr.points_), [0.25, 0.5, 0.75], rtol=1e-2)
        True
        >>> lr.score(X[1::2],y[1::2])
        0.7498

        References:
            .. footbibliography::

    """

    def __init__(
        self,
        max_features: int = 5,
        penalty: Literal["l1", "l2", "elasticnet", None] = None,
        C: float = 1,
        solver: Solver = 'lbfgs',
        max_iter: int = 100,
    ) -> None:

        self.max_features = max_features
        self.penalty = penalty
        self.C = C
        self.solver = solver
        self.max_iter = max_iter

    def fit(  # noqa: D102, WPS210
        self,
        X: FDataGrid,
        y: NDArrayAny,
    ) -> LogisticRegression:

        X, classes, y_ind = self._argcheck_X_y(X, y)

        self.classes_ = classes

        n_samples = len(y)
        n_features = len(X.grid_points[0])

        selected_indexes = np.empty(self.max_features, dtype=np.int_)
        selected_values = np.empty((n_samples, self.max_features))

        likelihood_curves_data = np.empty(
            (self.max_features, n_features),
        )

        penalty = 'none' if self.penalty is None else self.penalty

        # multivariate logistic regression
        mvlr = mvLogisticRegression(
            penalty=penalty,
            C=self.C,
            solver=self.solver,
            max_iter=self.max_iter,
        )

        max_features = min(self.max_features, len(X.grid_points[0]))

        last_max_likelihood = -np.inf

        for n_selected in range(max_features):
            for t in range(n_features):

                selected_values[:, n_selected] = X.data_matrix[:, t, 0]
                mvlr.fit(selected_values[:, :n_selected + 1], y_ind)

                # log-likelihood function at t
                with np.errstate(divide='ignore'):
                    log_probs = mvlr.predict_log_proba(
                        selected_values[:, :n_selected + 1],
                    )

                log_probs = np.concatenate(
                    (log_probs[y_ind == 0, 0], log_probs[y_ind == 1, 1]),
                )
                likelihood_curves_data[n_selected, t] = np.mean(log_probs)

            tmax = np.argmax(likelihood_curves_data[n_selected])
            max_likelihood = likelihood_curves_data[n_selected, tmax]
            if max_likelihood == last_max_likelihood:
                # This does not improve
                selected_indexes = selected_indexes[:n_selected]
                selected_values = selected_values[:, :n_selected]
                likelihood_curves_data = likelihood_curves_data[
                    :n_selected,
                    n_features - 1,
                ]
                break

            last_max_likelihood = max_likelihood

            selected_indexes[n_selected] = tmax
            selected_values[:, n_selected] = X.data_matrix[:, tmax, 0]

        # fit for the complete set of points
        mvlr.fit(selected_values, y)

        self.coef_ = mvlr.coef_
        self.intercept_ = mvlr.intercept_
        self._likelihood_curves = FDataGrid(
            likelihood_curves_data,
            grid_points=X.grid_points,
        )
        self._mvlr = mvlr

        self._selected_indexes = selected_indexes
        self.points_ = X.grid_points[0][selected_indexes]

        return self

    def predict(self, X: FDataGrid) -> NDArrayInt:  # noqa: D102
        check_is_fitted(self)
        return self._wrapper(self._mvlr.predict, X)

    def predict_log_proba(self, X: FDataGrid) -> NDArrayInt:  # noqa: D102
        check_is_fitted(self)
        return self._wrapper(self._mvlr.predict_log_proba, X)

    def predict_proba(self, X: FDataGrid) -> NDArrayInt:  # noqa: D102
        check_is_fitted(self)
        return self._wrapper(self._mvlr.predict_proba, X)

    def _argcheck_X(  # noqa: N802
        self,
        X: FDataGrid,
    ) -> FDataGrid:

        if X.dim_domain > 1:
            raise ValueError(
                f'The dimension of the domain has to be one'
                f'; got {X.dim_domain} dimensions',
            )

        return X

    def _argcheck_X_y(  # noqa: N802
        self,
        X: FDataGrid,
        y: NDArrayAny,
    ) -> Tuple[FDataGrid, NDArrayAny, NDArrayAny]:

        X = self._argcheck_X(X)

        classes, y_ind = _classifier_get_classes(y)

        if classes.size > 2:
            raise ValueError(
                f'The number of classes has to be two'
                f'; got {classes.size} classes',
            )

        if (len(y) != len(X)):
            raise ValueError(
                "The number of samples on independent variables"
                " and classes should be the same",
            )

        return (X, classes, y_ind)

    def _wrapper(
        self,
        method: Callable[[NDArrayAny], NDArrayAny],
        X: FDataGrid,
    ) -> NDArrayAny:
        """Wrap multivariate logistic regression method.

        This function transforms functional data in order to pass
        them to a multivariate logistic regression method.

        .. warning::
            This function can't be called before fit.
        """
        X = self._argcheck_X(X)

        x_mv = X.data_matrix[:, self._selected_indexes, 0]

        return method(x_mv)



"""Imputation of missing values."""
from __future__ import annotations

import importlib
from typing import TYPE_CHECKING, Any

import lazy_loader as lazy

__getattr__, __dir__, __all__ = lazy.attach(
    __name__,
    submod_attrs={
        "_interpolate": ["MissingValuesInterpolation"],
    },
)

if TYPE_CHECKING:
    from ._interpolate import (
        MissingValuesInterpolation as MissingValuesInterpolation
    )



"""NumPy aliases for compatibility."""

from typing import Any, Union

import numpy as np

try:
    from numpy.typing import ArrayLike as ArrayLike  # noqa: WPS113
except ImportError:
    ArrayLike = np.ndarray  # type:ignore[misc] # noqa: WPS440

try:  # noqa: WPS229
    from numpy.typing import NDArray
    NDArrayAny = NDArray[Any]
    NDArrayInt = NDArray[np.int_]
    NDArrayFloat = NDArray[np.float_]
    NDArrayReal = NDArray[Union[np.float_, np.int_]]
    NDArrayBool = NDArray[np.bool_]
    NDArrayStr = NDArray[np.str_]
    NDArrayObject = NDArray[np.object_]
except ImportError:
    NDArray = np.ndarray  # type:ignore[misc] # noqa: WPS440
    NDArrayAny = np.ndarray  # type:ignore[misc]
    NDArrayInt = np.ndarray  # type:ignore[misc]
    NDArrayFloat = np.ndarray  # type:ignore[misc]
    NDArrayReal = np.ndarray  # type:ignore[misc]
    NDArrayBool = np.ndarray  # type:ignore[misc]
    NDArrayStr = np.ndarray  # type:ignore[misc]
    NDArrayObject = np.ndarray  # type:ignore[misc]



"""Machine learning methods for functional data."""
import lazy_loader as lazy

__getattr__, __dir__, __all__ = lazy.attach(
    __name__,
    submodules=[
        "classification",
        "clustering",
        "regression",
    ],
)



from typing import TYPE_CHECKING

import lazy_loader as lazy

__getattr__, __dir__, __all__ = lazy.attach(
    __name__,
    submod_attrs={
        "_real_datasets": [
            "fdata_constructor",
            "fetch_aemet",
            "fetch_cran",
            "fetch_gait",
            "fetch_growth",
            "fetch_handwriting",
            "fetch_mco",
            "fetch_medflies",
            "fetch_nox",
            "fetch_octane",
            "fetch_phoneme",
            "fetch_tecator",
            "fetch_ucr",
            "fetch_weather",
        ],
        "_samples_generators": [
            "make_gaussian",
            "make_gaussian_process",
            "make_multimodal_landmarks",
            "make_multimodal_samples",
            "make_random_warping",
            "make_sinusoidal_process",
        ],
    },
)

if TYPE_CHECKING:
    from ._real_datasets import (
        fdata_constructor as fdata_constructor,
        fetch_aemet as fetch_aemet,
        fetch_cran as fetch_cran,
        fetch_gait as fetch_gait,
        fetch_growth as fetch_growth,
        fetch_handwriting as fetch_handwriting,
        fetch_mco as fetch_mco,
        fetch_medflies as fetch_medflies,
        fetch_nox as fetch_nox,
        fetch_octane as fetch_octane,
        fetch_phoneme as fetch_phoneme,
        fetch_tecator as fetch_tecator,
        fetch_ucr as fetch_ucr,
        fetch_weather as fetch_weather,
    )
    from ._samples_generators import (
        make_gaussian as make_gaussian,
        make_gaussian_process as make_gaussian_process,
        make_multimodal_landmarks as make_multimodal_landmarks,
        make_multimodal_samples as make_multimodal_samples,
        make_random_warping as make_random_warping,
        make_sinusoidal_process as make_sinusoidal_process,
    )
